# pylint: skip-file
try:
    from unittest import mock
except ImportError:
    import mock

import os
import six
import tempfile
import unittest
from unittest.mock import patch
from decimal import Decimal
from textwrap import dedent
from io import StringIO

from blivet import errors
from blivet import util
from blivet.size import Size


class MiscTest(unittest.TestCase):

    # Disable this warning, which will only be triggered on python3.  For
    # python2, the default is False.
    long_message = True      # pylint: disable=pointless-class-attribute-override

    def test_power_of_two(self):
        self.assertFalse(util.power_of_two(None))
        self.assertFalse(util.power_of_two("not a number"))
        self.assertFalse(util.power_of_two(Decimal("2.2")))
        self.assertFalse(util.power_of_two(-1))
        self.assertFalse(util.power_of_two(0))
        self.assertFalse(util.power_of_two(1))
        for i in range(1, 60, 5):
            self.assertTrue(util.power_of_two(2 ** i), msg=i)
            self.assertFalse(util.power_of_two(2 ** i + 1), msg=i)
            self.assertFalse(util.power_of_two(2 ** i - 1), msg=i)

    def test_dedup_list(self):
        # no duplicates, no change
        self.assertEqual([1, 2, 3, 4], util.dedup_list([1, 2, 3, 4]))
        # empty list no issue
        self.assertEqual([], util.dedup_list([]))

        # real deduplication
        self.assertEqual([1, 2, 3, 4, 5, 6], util.dedup_list([1, 2, 3, 4, 2, 2, 2, 1, 3, 5, 3, 6, 6, 2, 3, 1, 5]))

    def test_detect_virt(self):
        in_virt = not util.run_program(["systemd-detect-virt", "--vm"])
        self.assertEqual(util.detect_virt(), in_virt)


class TestDefaultNamedtuple(unittest.TestCase):
    def test_default_namedtuple(self):
        TestTuple = util.default_namedtuple("TestTuple", ["x", "y", ("z", 5), "w"])
        dnt = TestTuple(1, 2, 3, 6)
        self.assertEqual(dnt.x, 1)
        self.assertEqual(dnt.y, 2)
        self.assertEqual(dnt.z, 3)
        self.assertEqual(dnt.w, 6)
        self.assertEqual(dnt, (1, 2, 3, 6))

        dnt = TestTuple(1, 2, 3, w=6)
        self.assertEqual(dnt.x, 1)
        self.assertEqual(dnt.y, 2)
        self.assertEqual(dnt.z, 3)
        self.assertEqual(dnt.w, 6)
        self.assertEqual(dnt, (1, 2, 3, 6))

        dnt = TestTuple(z=3, x=2, w=1, y=5)
        self.assertEqual(dnt, (2, 5, 3, 1))

        dnt = TestTuple()
        self.assertEqual(dnt, (None, None, 5, None))


class Test(object):
    def __init__(self, s):
        self._s = s

    @property
    def s(self):
        return self._s

    @property
    def ok(self):
        return True

    @property
    def nok(self):
        return False

    @util.requires_property("s", "hi")
    def say_hi(self):
        return self.s + ", guys!"

    @property
    @util.requires_property("s", "hi")
    def hi(self):
        return self.s

    @property
    @util.requires_property("ok")
    def good_news(self):
        return "Everything okay!"

    @property
    @util.requires_property("nok")
    def bad_news(self):
        return "Nothing okay!"


class TestRequiresProperty(unittest.TestCase):
    def test_requires_property(self):
        t = Test("hi")

        self.assertEqual(t.s, "hi")
        self.assertEqual(t.say_hi(), "hi, guys!")
        self.assertEqual(t.hi, "hi")

        t = Test("hello")
        self.assertEqual(t.s, "hello")
        with self.assertRaises(ValueError):
            t.say_hi()
        with self.assertRaises(ValueError):
            t.hi()

        self.assertEqual(t.good_news, "Everything okay!")

        with self.assertRaises(ValueError):
            t.bad_news


class TestDependencyGuard(util.DependencyGuard):
    error_msg = "test dep not satisfied"

    def _check_avail(self):
        return False


_requires_something = TestDependencyGuard()


class DependencyGuardTestCase(unittest.TestCase):
    @_requires_something(critical=True)
    def _test_dependency_guard_critical(self):
        return True

    @_requires_something(critical=False)
    def _test_dependency_guard_non_critical(self):
        return True

    def test_dependency_guard(self):
        guard = TestDependencyGuard()
        if six.PY3:
            with self.assertLogs("blivet", level="WARNING") as cm:
                self.assertEqual(self._test_dependency_guard_non_critical(), None)
            self.assertTrue(TestDependencyGuard.error_msg in "\n".join(cm.output))
        else:
            self.assertEqual(self._test_dependency_guard_non_critical(), None)

        with self.assertRaises(errors.DependencyError):
            self._test_dependency_guard_critical()

        with mock.patch.object(_requires_something, '_check_avail', return_value=True):
            self.assertEqual(self._test_dependency_guard_non_critical(), True)
            self.assertEqual(self._test_dependency_guard_critical(), True)


class GetSysfsAttrTestCase(unittest.TestCase):

    def test_get_sysfs_attr(self):

        with tempfile.TemporaryDirectory() as sysfs:
            model_file = os.path.join(sysfs, "model")
            with open(model_file, "w") as f:
                f.write("test model\n")

            model = util.get_sysfs_attr(sysfs, "model")
            self.assertEqual(model, "test model")

            # now with some invalid byte in the model
            with open(model_file, "wb") as f:
                f.write(b"test model\xef\n")

            # the unicode replacement character (U+FFFD) should be used instead
            model = util.get_sysfs_attr(sysfs, "model")
            self.assertEqual(model, "test model\ufffd")


class GetKernelModuleParameterTestCase(unittest.TestCase):

    def test_nonexisting_kernel_module(self):
        self.assertIsNone(util.get_kernel_module_parameter("unknown_module", "unknown_parameter"))

    def test_get_kernel_module_parameter_value(self):
        with mock.patch('blivet.util.open', mock.mock_open(read_data='value\n')):
            value = util.get_kernel_module_parameter("module", "parameter")
        self.assertEqual(value, "value")


class MemoryTests(unittest.TestCase):

    MEMINFO = dedent(
        """MemTotal:       32526648 kB
           MemFree:         8196560 kB
           MemAvailable:   21189232 kB
           Buffers:            4012 kB
           Cached:         13974708 kB
           SwapCached:            0 kB
           Active:          4934172 kB
           Inactive:       17128972 kB
           Active(anon):       7184 kB
           Inactive(anon):  9202192 kB
           Active(file):    4926988 kB
           Inactive(file):  7926780 kB
           Unevictable:     1009932 kB
           Mlocked:             152 kB
           SwapTotal:       8388604 kB
           SwapFree:        8388604 kB
           Zswap:                 0 kB
           Zswapped:              0 kB
           Dirty:              4508 kB
           Writeback:             0 kB
           AnonPages:       9094440 kB
           Mapped:          1224920 kB
           Shmem:           1124952 kB
           KReclaimable:     605048 kB
           Slab:             969324 kB
           SReclaimable:     605048 kB
           SUnreclaim:       364276 kB
           KernelStack:       36672 kB
           PageTables:        85696 kB
           NFS_Unstable:          0 kB
           Bounce:                0 kB
           WritebackTmp:          0 kB
           CommitLimit:    24651928 kB
           Committed_AS:   19177064 kB
           VmallocTotal:   34359738367 kB
           VmallocUsed:      102060 kB
           VmallocChunk:          0 kB
           Percpu:            11392 kB
           HardwareCorrupted:     0 kB
           AnonHugePages:         0 kB
           ShmemHugePages:        0 kB
           ShmemPmdMapped:        0 kB
           FileHugePages:         0 kB
           FilePmdMapped:         0 kB
           CmaTotal:              0 kB
           CmaFree:               0 kB
           HugePages_Total:       0
           HugePages_Free:        0
           HugePages_Rsvd:        0
           HugePages_Surp:        0
           Hugepagesize:       2048 kB
           Hugetlb:               0 kB
           DirectMap4k:      829716 kB
           DirectMap2M:    27138048 kB
           DirectMap1G:     6291456 kB
        """
    )

    @patch("blivet.util.open")
    def test_total_memory_real(self, open_mock):
        """Test total_memory with real data"""
        open_mock.return_value = StringIO(self.MEMINFO)
        assert util.total_memory() == Size("32657720.0 KiB")
        open_mock.assert_called_once_with("/proc/meminfo", "r")

    @patch("blivet.util.open")
    def test_total_memory_missing(self, open_mock):
        """Test total_memory with missing value"""
        missing = self.MEMINFO.replace("MemTotal", "Nonsense")
        open_mock.return_value = StringIO(missing)
        with self.assertRaises(RuntimeError):
            util.total_memory()
        open_mock.assert_called_once_with("/proc/meminfo", "r")

    @patch("blivet.util.open")
    def test_total_memory_not_number(self, open_mock):
        """Test total_memory with bad format"""
        missing = self.MEMINFO.replace(
            "MemTotal:       32526648 kB",
            "MemTotal:       nonsense kB"
        )
        open_mock.return_value = StringIO(missing)
        with self.assertRaises(RuntimeError):
            util.total_memory()

        malformed = self.MEMINFO.replace(
            "MemTotal:       32526648 kB",
            "MemTotal:       32526648 kB as of right now"
        )
        open_mock.return_value = StringIO(malformed)
        with self.assertRaises(RuntimeError):
            util.total_memory()

    @patch("blivet.util.open")
    def test_total_memory_calculations(self, open_mock):
        """Test total_memory calculates correctly."""
        open_mock.return_value = StringIO("MemTotal: 1024 kB")
        assert util.total_memory() == Size("132096.0 KiB")

        open_mock.return_value = StringIO("MemTotal: 65536 kB")
        assert util.total_memory() == Size("196608.0 KiB")

        open_mock.return_value = StringIO("MemTotal: 10000000 kB")
        assert util.total_memory() == Size("10131072.0 KiB")



# lukstasks.py
# Tasks for a LUKS format.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

from .. import util

from ..devicelibs import crypto
from ..errors import LUKSError
from ..devices.lib import LINUX_SECTOR_SIZE

from . import availability
from . import task
from . import dfresize


class LUKSSize(task.BasicApplication):
    """ Obtain information about the size of a LUKS format. """

    ext = availability.BLOCKDEV_DM_PLUGIN

    description = "size of a luks device"

    def __init__(self, a_luks):
        """ Initializer.

            :param :class:`~.formats.luks.LUKS` a_luks: a LUKS format object
        """
        self.luks = a_luks

    def do_task(self):  # pylint: disable=arguments-differ
        """ Returns the size of the luks format.

            :returns: the size of the luks format
            :rtype: :class:`~.size.Size`
            :raises :class:`~.errors.LUKSError`: if size cannot be obtained
        """

        try:
            dm_dev = blockdev.dm.node_from_name(self.luks.map_name)
            blocks = int(util.get_sysfs_attr("/sys/block/%s" % dm_dev, "size") or "0")
        except blockdev.DMError as e:
            raise LUKSError(e)

        return blocks * LINUX_SECTOR_SIZE


class LUKSResize(task.BasicApplication, dfresize.DFResizeTask):
    """ Handle resize of LUKS device. """

    description = "resize luks device"

    ext = availability.BLOCKDEV_CRYPTO_PLUGIN

    # units for specifying new size
    unit = crypto.SECTOR_SIZE

    def __init__(self, a_luks):
        """ Initializer.

            :param :class:`~.formats.luks.LUKS` a_luks: a LUKS format object
        """
        self.luks = a_luks

    def do_task(self):  # pylint: disable=arguments-differ
        """ Resizes the LUKS format. """
        try:
            if self.luks.luks_version == "luks2":
                if self.luks._LUKS__passphrase:
                    context = blockdev.CryptoKeyslotContext(passphrase=self.luks._LUKS__passphrase)
                elif self.luks._key_file:
                    context = blockdev.CryptoKeyslotContext(keyfile=self.luks._key_file)
                else:
                    # context for resize can be NULL -- this means the key is already in the keyring
                    context = None
                blockdev.crypto.luks_resize(self.luks.map_name, self.luks.target_size.convert_to(self.unit),
                                            context=context)
            else:
                blockdev.crypto.luks_resize(self.luks.map_name, self.luks.target_size.convert_to(self.unit))
        except blockdev.CryptoError as e:
            raise LUKSError(e)



# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>

import unittest
import unittest.mock
import tempfile
import warnings
import polib
import os

from translation_canary.translated.test_markup import test_markup
from translation_canary.translated.test_percentage import test_percentage
from translation_canary.translated.test_usability import test_usability, test_msgfmt

from translation_canary.translated import testFile, testSourceTree

# convert a polib.POFile into a NamedTemporaryFile
def pofile(poobj):
    f = tempfile.NamedTemporaryFile(suffix='.po')
    poobj.save(f.name)
    return f

# convenience function for creating a single-entry mofile
def pofile_from_entry(*args, **kwargs):
    poobj = polib.POFile()
    poobj.metadata["Content-Type"] = "text/plain; charset=UTF-8"
    poobj.append(polib.POEntry(*args, **kwargs))
    return pofile(poobj)

class TestMarkup(unittest.TestCase):
    # I know, pylint, that's the point
    # pylint: disable=invalid-markup

    def test_ok(self):
        # no markup
        with pofile_from_entry(msgid="test string", msgstr="estay ingstray") as p:
            test_markup(p.name)

        # matching markup
        with pofile_from_entry(msgid="<b>bold</b> string", msgstr="<b>oldbay</b> ingstray") as p:
            test_markup(p.name)

        # matching plural
        with pofile_from_entry(msgid="%d <b>bold</b> string", msgid_plural="%d <b>bold</b> strings",
                msgstr_plural={0: "%d <b>oldbay</b> ingstray", 1: "%d <b>oldbay</b> instrays"}) as p:
            test_markup(p.name)

    def test_missing(self):
        with pofile_from_entry(msgid="<b>bold</b> string", msgstr="oldbay ingstray") as p:
            self.assertRaises(AssertionError, test_markup, p.name)

    def test_mismatch(self):
        with pofile_from_entry(msgid="<b>bold</b> string", msgstr="<i>oldbay</i> ingstray") as p:
            self.assertRaises(AssertionError, test_markup, p.name)

    def test_typo(self):
        with pofile_from_entry(msgid="<b>bold</b> string", msgstr="<boldbay</b> ingstray") as p:
            self.assertRaises(AssertionError, test_markup, p.name)

    def test_mismatch_plural(self):
        with pofile_from_entry(msgid="%d <b>bold</b> string", msgid_plural="%d <b>bold</b> strings",
                msgstr_plural={0: "%d <b>olbday</b> ingstray", 1: "%d oldbay ingstrays"}) as p:
            self.assertRaises(AssertionError, test_markup, p.name)

    def test_invalid(self):
        # Tags themselves are valid, but the XML is not
        with pofile_from_entry(msgid="<b>bold</b> string", msgstr="<b/>oldbay</b> ingstray") as p:
            self.assertRaises(AssertionError, test_markup, p.name)

class TestPercentage(unittest.TestCase):
    def setUp(self):
        # Convert warnings into exceptions to make them easier to test for
        warnings.simplefilter("error")
        # polib throws a DeprecationWarnings so ignore that
        warnings.simplefilter("default", DeprecationWarning)

    def tearDown(self):
        warnings.resetwarnings()

    def test_ok(self):
        # 100%
        with pofile_from_entry(msgid="test string", msgstr="estay ingstray") as p:
            test_percentage(p.name)

    def test_not_ok(self):
        # 0%
        with pofile_from_entry(msgid="test string", msgstr="") as p:
            self.assertRaises(Warning, test_percentage, p.name)

class TestUsability(unittest.TestCase):
    def test_ok(self):
        # what lt's Plural-Forms is supposed to look like
        poobj = polib.POFile()
        poobj.metadata["Plural-Forms"] = "nplurals=3; plural=(n%10==1 && n%100!=11 ? 0 : n%10>=2 && (n%100<10 || n%100>=20) ? 1 : 2)\n"

        with pofile(poobj) as p:
            test_usability(p.name)

    def test_busted_plural_forms(self):
        # https://bugzilla.redhat.com/show_bug.cgi?id=1283599
        poobj = polib.POFile()
        poobj.metadata["Plural-Forms"] = "nplurals=3; plural=(n%10==1 && n%100!=11 ? 0 : n%10>=2 && (n%100<10 or n%100>=20) ? 1 : 2)\n"

        with pofile(poobj) as p:
            self.assertRaises(Exception, test_usability, p.name)

class TestMsgFmt(unittest.TestCase):
    def test_ok(self):
        with pofile_from_entry(msgid="test string", msgstr="estay ingstray") as p:
            test_msgfmt(p.name)

    # Test a few cases that msgfmt will catch
    def test_busted_newlines(self):
        with pofile_from_entry(msgid="multi\nline\nstring", msgstr="ultimay\ninelay\ningstray\n") as p:
            self.assertRaises(AssertionError, test_msgfmt, p.name)

    def test_busted_format(self):
        with pofile_from_entry(msgid="test %(type)s", msgstr="estay", flags=["python-format"]) as p:
            self.assertRaises(AssertionError, test_msgfmt, p.name)

    def test_translated_format(self):
        with pofile_from_entry(msgid="test %(type)s", msgstr="estay %(ypetay)", flags=["python-format"]) as p:
            self.assertRaises(AssertionError, test_msgfmt, p.name)

# Test the top-level functions
# fake tests for testing with
def _true_test(_p):
    pass

def _false_test(_p):
    raise AssertionError("no good")

def _picky_test(p):
    if os.path.basename(p).startswith("p"):
        raise AssertionError("I don't like this one")

class TestTestFile(unittest.TestCase):
    @unittest.mock.patch("translation_canary.translated._tests", [_true_test])
    def test_success(self):
        with pofile_from_entry() as p:
            self.assertTrue(testFile(p.name))

    @unittest.mock.patch("translation_canary.translated._tests", [_false_test])
    def test_failure(self):
        with pofile_from_entry() as p:
            self.assertFalse(testFile(p.name))

    @unittest.mock.patch("translation_canary.translated._tests", [_true_test])
    def test_release_mode_success(self):
        with pofile_from_entry() as p:
            self.assertTrue(testFile(p.name, releaseMode=True))
            self.assertTrue(os.path.exists(p.name))

    @unittest.mock.patch("translation_canary.translated._tests", [_false_test])
    def test_release_mode_failure(self):
        p = tempfile.NamedTemporaryFile(suffix='.po', delete=False)
        try:
            # testFile should return True but delete the file
            self.assertTrue(testFile(p.name, releaseMode=True))
            self.assertFalse(os.path.exists(p.name))
        finally:
            try:
                p.close()
                os.unlink(p.name)
            except FileNotFoundError:
                pass

    @unittest.mock.patch("translation_canary.translated._tests", [_false_test])
    def test_release_mode_failure_with_lingua(self):
        with tempfile.TemporaryDirectory() as d:
            open(os.path.join(d, "test.po"), "w").close()
            with open(os.path.join(d, "LINGUAS"), "w") as linguas:
                linguas.write("test other\n")

            # Check that test.po is removed and test is removed from LINGUAS
            self.assertTrue(testFile(os.path.join(d, "test.po"), releaseMode=True))
            self.assertFalse(os.path.exists(os.path.join(d, "test.po")))

            with open(os.path.join(d, "LINGUAS"), "r")  as linguas:
                self.assertEqual(linguas.read().strip(), "other")

    @unittest.mock.patch("translation_canary.translated._tests", [_false_test])
    def test_release_mode_failure_with_lingua_no_modify(self):
        with tempfile.TemporaryDirectory() as d:
            open(os.path.join(d, "test.po"), "w").close()
            with open(os.path.join(d, "LINGUAS"), "w") as linguas:
                linguas.write("test other\n")

            # Check that test.po is removed and test is *not* removed from LINGUAS
            self.assertTrue(testFile(os.path.join(d, "test.po"), releaseMode=True, modifyLinguas=False))
            self.assertFalse(os.path.exists(os.path.join(d, "test.po")))

            with open(os.path.join(d, "LINGUAS"), "r") as linguas:
                self.assertEqual(linguas.read().strip(), "test other")

@unittest.mock.patch("translation_canary.translated._tests", [_picky_test])
class TestTestSourceTree(unittest.TestCase):
    def setUp(self):
        self.tmpobj = tempfile.TemporaryDirectory()
        self.tmpdir = self.tmpobj.name

    def tearDown(self):
        self.tmpobj.cleanup()

    def test_success(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "ab.po"), "w").close()
        self.assertTrue(testSourceTree(self.tmpdir))

    def test_some_failure(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "pa.po"), "w").close()
        self.assertFalse(testSourceTree(self.tmpdir))

    def test_all_failure(self):
        open(os.path.join(self.tmpdir, "pa.po"), "w").close()
        open(os.path.join(self.tmpdir, "pb.po"), "w").close()
        self.assertFalse(testSourceTree(self.tmpdir))

    def test_release_mode_success(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "ab.po"), "w").close()
        self.assertTrue(testSourceTree(self.tmpdir, releaseMode=True))

    def test_release_mode_failure(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "pa.po"), "w").close()
        self.assertTrue(testSourceTree(self.tmpdir, releaseMode=True))
        self.assertTrue(os.path.exists(os.path.join(self.tmpdir, "aa.po")))
        self.assertFalse(os.path.exists(os.path.join(self.tmpdir, "pa.po")))

    def test_release_mode_failure_with_lingua(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "pa.po"), "w").close()
        with open(os.path.join(self.tmpdir, "LINGUAS"), "w") as l:
            l.write("aa pa\n")

        self.assertTrue(testSourceTree(self.tmpdir, releaseMode=True))
        self.assertTrue(os.path.exists(os.path.join(self.tmpdir, "aa.po")))
        self.assertFalse(os.path.exists(os.path.join(self.tmpdir, "pa.po")))

        with open(os.path.join(self.tmpdir, "LINGUAS")) as l:
            self.assertEqual(l.read().strip(), "aa")

    def test_release_mode_failure_with_lingua_no_modify(self):
        open(os.path.join(self.tmpdir, "aa.po"), "w").close()
        open(os.path.join(self.tmpdir, "pa.po"), "w").close()

        with open(os.path.join(self.tmpdir, "LINGUAS"), "w") as l:
            l.write("aa pa\n")

        self.assertTrue(testSourceTree(self.tmpdir, releaseMode=True, modifyLinguas=False))
        self.assertTrue(os.path.exists(os.path.join(self.tmpdir, "aa.po")))
        self.assertFalse(os.path.exists(os.path.join(self.tmpdir, "pa.po")))

        with open(os.path.join(self.tmpdir, "LINGUAS")) as l:
            self.assertEqual(l.read().strip(), "aa pa")



# devices/md.py
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import math
import os
import six
import time

from six.moves import reduce

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

from ..devicelibs import mdraid, raid

from .. import errors
from ..formats import DeviceFormat
from .. import util
from ..static_data import pvs_info
from ..storage_log import log_method_call
from .. import udev
from ..size import Size
from ..tasks import availability

import logging
log = logging.getLogger("blivet")

from .storage import StorageDevice
from .container import ContainerDevice
from .raid import RaidDevice
from .lib import get_majors_by_device_type

MD_MAJORS = get_majors_by_device_type("md")


class MDRaidArrayDevice(ContainerDevice, RaidDevice):

    """ An mdraid (Linux RAID) device. """
    _type = "mdarray"
    _packages = ["mdadm"]
    _dev_dir = "/dev/md"
    _format_class_name = property(lambda s: "mdmember")
    _format_uuid_attr = property(lambda s: "md_uuid")
    _external_dependencies = [availability.BLOCKDEV_MDRAID_PLUGIN]

    def __init__(self, name, level=None, major=None, minor=None, size=None,
                 member_devices=None, total_devices=None,
                 uuid=None, fmt=None, exists=False, metadata_version=None,
                 parents=None, sysfs_path='', chunk_size=None):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword sysfs_path: sysfs device path
            :type sysfs_path: str
            :keyword chunk_size: chunk size for the device
            :type chunk_size: :class:`~.size.Size`
            :keyword uuid: the device UUID
            :type uuid: str

            :keyword level: the device's RAID level
            :type level: any valid RAID level descriptor
            :keyword int member_devices: the number of active member devices
            :keyword int total_devices: the total number of member devices
            :keyword metadata_version: the version of the device's md metadata
            :type metadata_version: str (eg: "0.90")
            :keyword minor: the device minor (obsolete?)
            :type minor: int

            .. note::

                An instance of this class whose :attr:`exists` attribute is
                True and whose parent/member devices are all partitionable is
                also considered to be partitionable.

            .. note::

                An instance of this class whose :attr:`exists` attribute is
                True and whose parent/member devices are all disks is also
                treated like a disk.

        """
        # pylint: disable=unused-argument

        # These attributes are used by _add_parent, so they must be initialized
        # prior to instantiating the superclass.
        self._member_devices = 0     # the number of active (non-spare) members
        self._total_devices = 0      # the total number of members

        # avoid attribute-defined-outside-init pylint warning
        self._level = None

        super(MDRaidArrayDevice, self).__init__(name, uuid=uuid,
                                                exists=exists, size=size,
                                                parents=parents,
                                                sysfs_path=sysfs_path)

        try:
            self.level = level
        except errors.DeviceError as e:
            # Could not set the level, so set loose the parents that were
            # added in superclass constructor.
            for dev in self.parents:
                dev.remove_child(self)
            raise e

        self.uuid = uuid
        self._total_devices = util.numeric_type(total_devices)
        self.member_devices = util.numeric_type(member_devices)

        if self.exists:
            self._chunk_size = self.read_chunk_size()
        else:
            self.chunk_size = chunk_size or Size(0)

        if not self.exists and not isinstance(metadata_version, str):
            self.metadata_version = "default"
        else:
            self.metadata_version = metadata_version

        self.format = fmt

        if self.parents and self.parents[0].type == "mdcontainer" and self.type != "mdbiosraidarray":
            raise errors.DeviceError("A device with mdcontainer member must be mdbiosraidarray.")

    @property
    def mdadm_format_uuid(self):
        """ This array's UUID, formatted for external use.

            :returns: the array's UUID in mdadm format, if available
            :rtype: str or NoneType
        """
        formatted_uuid = None

        if self.uuid is not None:
            try:
                formatted_uuid = blockdev.md.get_md_uuid(self.uuid)
            except blockdev.MDRaidError:
                pass

        return formatted_uuid

    @property
    def level(self):
        """ Return the raid level

            :returns: raid level value
            :rtype:   an object that represents a RAID level
        """
        return self._level

    @property
    def _levels(self):
        """ Allowed RAID level for this type of device."""
        return mdraid.raid_levels

    @level.setter
    def level(self, value):
        """ Set the RAID level and enforce restrictions based on it.

            :param value: new raid level
            :param type:  object
            :raises :class:`~.errors.DeviceError`: if value does not describe
            a valid RAID level
            :returns:     None
        """
        try:
            level = self._get_level(value, self._levels)
        except ValueError as e:
            raise errors.DeviceError(e)

        self._level = level

    @property
    def sector_size(self):
        if not self.exists:
            # Least common multiple of parents' sector sizes
            return reduce(lambda a, b: a * b // math.gcd(a, b), (int(p.sector_size) for p in self.parents))

        return super(MDRaidArrayDevice, self).sector_size

    @property
    def chunk_size(self):
        if self._chunk_size == Size(0):
            if self.exists:
                return self.read_chunk_size()
            else:
                if self.level == raid.RAID1:
                    return self._chunk_size
                else:
                    return mdraid.MD_CHUNK_SIZE
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, newsize):
        if not isinstance(newsize, Size):
            raise ValueError("new chunk size must be of type Size")

        if newsize % Size("4 KiB") != Size(0):
            raise ValueError("new chunk size must be multiple of 4 KiB")

        if self.exists:
            raise ValueError("cannot set chunk size for an existing device")

        if self.level == raid.RAID1 and newsize != Size(0):
            raise ValueError("specifying chunk size is not allowed for raid1")

        self._chunk_size = newsize

    def read_chunk_size(self):
        log_method_call(self, exists=self.exists, path=self.path,
                        sysfs_path=self.sysfs_path)
        chunk_size = Size(0)
        if self.status:
            chunk_size = Size(util.get_sysfs_attr(self.sysfs_path, "md/chunk_size") or "0")

        return chunk_size

    @property
    def create_bitmap(self):
        """ Whether or not a bitmap should be created on the array.

            If the array is sufficiently small, a bitmap yields no benefit.

            If the array has no redundancy, a bitmap is just pointless.
        """
        try:
            return self.level.has_redundancy() and self.size >= Size(1000) and self.format.type != "swap"
        except errors.RaidError:
            # If has_redundancy() raises an exception then this device has
            # a level for which the redundancy question is meaningless. In
            # that case, creating a write-intent bitmap would be a meaningless
            # action.
            return False

    def get_superblock_size(self, raw_array_size):
        """Estimate the superblock size for a member of an array,
           given the total available memory for this array and raid level.

           :param raw_array_size: total available for this array and level
           :type raw_array_size: :class:`~.size.Size`
           :returns: estimated superblock size
           :rtype: :class:`~.size.Size`
        """
        try:
            return blockdev.md.get_superblock_size(raw_array_size,
                                                   version=self.metadata_version)
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)

    @property
    def size(self):
        """Returns the actual or estimated size depending on whether or
           not the array exists.
        """
        if not self.exists or not self.media_present:
            try:
                size = self.level.get_size([d.size for d in self.members],
                                           self.member_devices,
                                           self.chunk_size,
                                           self.get_superblock_size)
            except (blockdev.MDRaidError, errors.RaidError) as e:
                log.info("could not calculate size of device %s for raid level %s: %s", self.name, self.level, e)
                size = Size(0)
            log.debug("non-existent RAID %s size == %s", self.level, size)
        else:
            size = self.current_size
            log.debug("existing RAID %s size == %s", self.level, size)

        return size

    def update_size(self, newsize=None):
        # container size is determined by the member disks, so there is nothing
        # to update in that case
        if self.type != "mdcontainer":
            # pylint: disable=bad-super-call
            super(ContainerDevice, self).update_size(newsize=newsize)

    @property
    def description(self):
        levelstr = self.level.nick if self.level.nick else self.level.name
        return "MDRAID set (%s)" % levelstr

    def __repr__(self):
        s = StorageDevice.__repr__(self)
        s += ("  level = %(level)s  spares = %(spares)s\n"
              "  members = %(member_devices)s\n"
              "  total devices = %(total_devices)s"
              "  metadata version = %(metadata_version)s" %
              {"level": self.level, "spares": self.spares,
               "member_devices": self.member_devices,
               "total_devices": self.total_devices,
               "metadata_version": self.metadata_version})
        return s

    @property
    def dict(self):
        d = super(MDRaidArrayDevice, self).dict
        d.update({"level": str(self.level),
                  "spares": self.spares, "member_devices": self.member_devices,
                  "total_devices": self.total_devices,
                  "metadata_version": self.metadata_version})
        return d

    @property
    def mdadm_conf_entry(self):
        """ This array's mdadm.conf entry. """
        uuid = self.mdadm_format_uuid
        if self.member_devices is None or not uuid:
            raise errors.DeviceError("array %s is not fully defined" % self.name)

        fmt = "ARRAY %s level=%s num-devices=%d UUID=%s\n"
        return fmt % (self.path, self.level, self.member_devices, uuid)

    @property
    def total_devices(self):
        """ Total number of devices in the array, including spares. """
        if not self.exists:
            return self._total_devices
        else:
            return len(self.parents)

    def _get_member_devices(self):
        return self._member_devices

    def _set_member_devices(self, number):
        if not isinstance(number, six.integer_types):
            raise ValueError("member_devices must be an integer")

        if not self.exists and number > self.total_devices:
            raise ValueError("member_devices cannot be greater than total_devices")
        self._member_devices = number

    member_devices = property(lambda d: d._get_member_devices(),
                              lambda d, m: d._set_member_devices(m),
                              doc="number of member devices")

    def _get_spares(self):
        spares = 0
        if self.member_devices is not None:
            if self.total_devices is not None and \
               self.total_devices > self.member_devices:
                spares = self.total_devices - self.member_devices
            elif self.total_devices is None:
                spares = self.member_devices
                self._total_devices = self.member_devices
        return spares

    def _set_spares(self, spares):
        max_spares = self.level.get_max_spares(len(self.parents))
        if spares > max_spares:
            log.debug("failed to set new spares value %d (max is %d)",
                      spares, max_spares)
            raise errors.DeviceError("new spares value is too large")

        if self.total_devices > spares:
            self.member_devices = self.total_devices - spares

    spares = property(_get_spares, _set_spares)

    def _add_parent(self, parent):
        super(MDRaidArrayDevice, self)._add_parent(parent)

        if self.status and parent.format.exists:
            # we always probe since the device may not be set up when we want
            # information about it
            self._size = self.current_size

        # These should be incremented when adding new parent devices except
        # during devicetree.populate. When detecting existing arrays we will
        # have gotten these values from udev and will use them to determine
        # whether we found all of the parents, so we shouldn't change them in
        # that case.
        if not parent.format.exists:
            self._total_devices += 1
            self.member_devices += 1

        # The new parent hasn't been added yet, so account for it explicitly.
        is_disk = self.is_disk and parent.is_disk
        for p in self.parents:
            p.format._hidden = is_disk

        parent.format._hidden = is_disk

    def _remove_parent(self, parent):
        error_msg = self._validate_parent_removal(self.level, parent)
        if error_msg:
            raise errors.DeviceError(error_msg)

        super(MDRaidArrayDevice, self)._remove_parent(parent)
        self.member_devices -= 1

    @property
    def _true_status_strings(self):
        """ Strings in state file for which status() should return True."""
        return ("clean", "active", "active-idle", "readonly", "read-auto")

    @property
    def status(self):
        """ This device's status.

            For now, this should return a boolean:
                True    the device is open and ready for use
                False   the device is not open
        """
        # check the status in sysfs
        status = False
        if not self.exists:
            return status

        if os.path.exists(self.path) and not self.sysfs_path:
            # the array has been activated from outside of blivet
            self.update_sysfs_path()

            # make sure the active array is the one we expect
            info = udev.get_device(self.sysfs_path)
            uuid = udev.device_get_md_uuid(info)
            if uuid and uuid != self.uuid:
                log.warning("md array %s is active, but has UUID %s -- not %s",
                            self.path, uuid, self.uuid)
                self.sysfs_path = ""
                return status

        state_file = "%s/md/array_state" % self.sysfs_path
        try:
            state = util.read_file(state_file).strip()
            if state in self._true_status_strings:
                status = True
        except OSError:
            status = False

        return status

    def member_status(self, member):
        if not (self.status and member.status):
            return

        member_name = os.path.basename(member.sysfs_path)
        path = "/sys/%s/md/dev-%s/state" % (self.sysfs_path, member_name)
        try:
            state = util.read_file(path).strip()
        except OSError:
            state = None

        return state

    @property
    def degraded(self):
        """ Return True if the array is running in degraded mode. """
        rc = False
        degraded_file = "%s/md/degraded" % self.sysfs_path
        if os.access(degraded_file, os.R_OK):
            val = util.read_file(degraded_file).strip()
            if val == "1":
                rc = True

        return rc

    @property
    def members(self):
        """ Returns this array's members.

            :rtype: list of :class:`StorageDevice`
        """
        return list(self.parents)

    @property
    def complete(self):
        """ An MDRaidArrayDevice is complete if it has at least as many
            component devices as its count of active devices.
        """
        return (self.member_devices <= len(self.members)) or not self.exists

    @property
    def bootable(self):
        return (self.level == raid.RAID1 and
                self.metadata_version in ("0.90", "1.0") and
                all(getattr(p, "bootable", False) for p in self.parents))

    def _post_setup(self):
        super(MDRaidArrayDevice, self)._post_setup()
        self.update_sysfs_path()

    def _setup(self, orig=False):
        """ Open, or set up, a device. """
        log_method_call(self, self.name, orig=orig, status=self.status,
                        controllable=self.controllable)
        disks = []
        for member in self.members:
            member.setup(orig=orig)
            disks.append(member.path)

        try:
            blockdev.md.activate(self.path, members=disks, uuid=self.mdadm_format_uuid)
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)

    def _post_teardown(self, recursive=False):
        super(MDRaidArrayDevice, self)._post_teardown(recursive=recursive)
        # mdadm reuses minors indiscriminantly when there is no mdadm.conf, so
        # we need to clear the sysfs path now so our status method continues to
        # give valid results
        self.sysfs_path = ''

        # make sure the /dev/mdXXX path is removed after stopping the array
        udev.settle()

    def teardown(self, recursive=None):
        """ Close, or tear down, a device. """
        log_method_call(self, self.name, status=self.status,
                        controllable=self.controllable)
        # we don't really care about the return value of _pre_teardown here.
        # see comment just above md_deactivate call
        self._pre_teardown(recursive=recursive)

        if self.is_disk:
            # treat arrays whose members are disks as partitionable disks
            return

        # We don't really care what the array's state is. If the device
        # file exists, we want to deactivate it. mdraid has too many
        # states.
        if self.exists and os.path.exists(self.path):
            try:
                blockdev.md.deactivate(self.path)
            except blockdev.MDRaidError as err:
                raise errors.MDRaidError(err)

        self._post_teardown(recursive=recursive)

    def pre_commit_fixup(self, current_fmt=False):
        """ Determine create parameters for this set """
        log_method_call(self, self.name)
        # UEFI firmware/bootloader cannot read 1.1 or 1.2 metadata arrays
        if getattr(self.format, "mountpoint", None) == "/boot/efi":
            self.metadata_version = "1.0"

    def _post_create(self):
        # this is critical since our status method requires a valid sysfs path
        self.exists = True  # this is needed to run update_sysfs_path
        self.update_sysfs_path()
        StorageDevice._post_create(self)

        # update our uuid attribute with the new array's UUID
        # XXX this won't work for containers since no UUID is reported for them
        try:
            info = blockdev.md.detail(self.path)
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)

        self.uuid = info.uuid
        for member in self.members:
            member.format.md_uuid = self.uuid

        def remove_stale_lvm():
            """ Remove any stale LVM metadata that pre-existed in a new array's on-disk footprint. """
            log.debug("waiting 5s for activation of stale lvm on new md array %s", self.path)
            time.sleep(5)
            udev.settle()
            pvs_info.drop_cache()
            pv_info = pvs_info.cache.get(self.path)
            if pv_info is None:
                return

            if pv_info.vg_uuid:
                log.info("removing stale LVM metadata found on %s", self.name)
                try:
                    blockdev.lvm.vgremove(pv_info.vg_name, extra={"--select": "vg_uuid=%s" % pv_info.vg_uuid})
                except blockdev.LVMError as e:
                    log.error("Failed to remove stale volume group from newly-created md array %s: %s",
                              self.path, str(e))
                    raise errors.MDRaidError(e)

            # lvm says it is a pv whether or not there is vg metadata, so wipe the pv signature
            try:
                blockdev.lvm.pvremove(self.path)
            except blockdev.LVMError as err:
                raise errors.MDRaidError(err)

        remove_stale_lvm()

        # remove any other stale metadata before proceeding
        DeviceFormat(device=self.path, exists=True).destroy()

    def _create(self):
        """ Create the device. """
        log_method_call(self, self.name, status=self.status)
        disks = [disk.path for disk in self.members]
        spares = len(self.members) - self.member_devices
        level = None
        if self.level:
            level = str(self.level)
        try:
            blockdev.md.create(self.path, level, disks, spares,
                               version=self.metadata_version,
                               bitmap="internal" if self.create_bitmap else None,
                               chunk_size=int(self.chunk_size))
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)
        udev.settle()

    def _remove(self, member):
        self.setup()
        # see if the device must be marked as failed before it can be removed
        fail = (self.member_status(member) == "in_sync")
        try:
            blockdev.md.remove(self.path, member.path, fail)
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)

    def _add(self, member):
        """ Add a member device to an array.

           :param str member: the member's path

           :raises: :class:`~.errors.MDRaidError`
        """
        self.setup()

        raid_devices = None
        try:
            if not self.level.has_redundancy():
                if self.level is not raid.Linear:
                    raid_devices = int(blockdev.md.detail(self.name).raid_devices) + 1
        except errors.RaidError:
            pass

        try:
            blockdev.md.add(self.path, member.path, raid_devs=raid_devices)
        except blockdev.MDRaidError as err:
            raise errors.MDRaidError(err)

    @property
    def format_args(self):
        format_args = []
        if self.format.type == "ext2":
            recommended_stride = self.level.get_recommended_stride(self.member_devices)
            if recommended_stride:
                format_args = ['-R', 'stride=%d' % recommended_stride]
        return format_args

    @property
    def model(self):
        return self.description

    @property
    def partitionable(self):
        return self.exists and self.parents and all(p.partitionable for p in self.members)

    @property
    def is_disk(self):
        return self.exists and self.parents and all(p.is_disk for p in self.members)

    def dracut_setup_args(self):
        return set(["rd.md.uuid=%s" % self.mdadm_format_uuid])

    def populate_ksdata(self, data):
        if self.is_disk:
            return

        super(MDRaidArrayDevice, self).populate_ksdata(data)
        data.level = self.level.name
        data.spares = self.spares
        data.members = ["raid.%d" % p.id for p in self.parents]
        data.preexist = self.exists
        data.device = self.name

        if not self.exists:
            # chunk size is meaningless on RAID1, so do not add our default value
            # to generated kickstart
            if self.level != raid.RAID1:
                data.chunk_size = self.chunk_size.convert_to("KiB")


class MDContainerDevice(MDRaidArrayDevice):

    _type = "mdcontainer"

    def __init__(self, name, **kwargs):
        kwargs['level'] = raid.Container
        super(MDContainerDevice, self).__init__(name, **kwargs)

    @property
    def _levels(self):
        return mdraid.MDRaidLevels(["container"])

    @property
    def description(self):
        return "BIOS RAID container"

    @property
    def mdadm_conf_entry(self):
        uuid = self.mdadm_format_uuid
        if not uuid:
            raise errors.DeviceError("array %s is not fully defined" % self.name)

        return "ARRAY %s UUID=%s\n" % (self.path, uuid)

    @property
    def _true_status_strings(self):
        return ("clean", "active", "active-idle", "readonly", "read-auto", "inactive")

    def teardown(self, recursive=None):
        log_method_call(self, self.name, status=self.status,
                        controllable=self.controllable)
        # we don't really care about the return value of _pre_teardown here.
        # see comment just above md_deactivate call
        self._pre_teardown(recursive=recursive)

        # Since BIOS RAID sets (containers in mdraid terminology) never change
        # there is no need to stop them and later restart them. Not stopping
        # (and thus also not starting) them also works around bug 523334
        return

    @property
    def media_present(self):
        # Containers should not get any format handling done
        # (the device node does not allow read / write calls)
        return False

    @property
    def is_disk(self):
        return False

    @property
    def partitionable(self):
        return False


class MDBiosRaidArrayDevice(MDRaidArrayDevice):

    _type = "mdbiosraidarray"
    _format_class_name = property(lambda s: None)
    _is_disk = True
    _partitionable = True

    def __init__(self, name, **kwargs):
        super(MDBiosRaidArrayDevice, self).__init__(name, **kwargs)

        # For container members probe size now, as we cannot determine it
        # when teared down.
        self._size = self.current_size

    @property
    def is_disk(self):
        # pylint: disable=bad-super-call
        # skip MDRaidArrayDevice and use the version in StorageDevice
        return super(MDRaidArrayDevice, self).is_disk

    @property
    def partitionable(self):
        # pylint: disable=bad-super-call
        # skip MDRaidArrayDevice and use the version in StorageDevice
        return super(MDRaidArrayDevice, self).partitionable

    @property
    def size(self):
        # For container members return probed size, as we cannot determine it
        # when teared down.
        return self._size

    @property
    def description(self):
        levelstr = self.level.nick if self.level.nick else self.level.name
        return "BIOS RAID set (%s)" % levelstr

    @property
    def mdadm_conf_entry(self):
        uuid = self.mdadm_format_uuid
        if not uuid:
            raise errors.DeviceError("array %s is not fully defined" % self.name)

        return "ARRAY %s UUID=%s\n" % (self.path, uuid)

    @property
    def members(self):
        # If the array is a BIOS RAID array then its unique parent
        # is a container and its actual member devices are the
        # container's parents.
        return self.parents[0].members

    @property
    def total_devices(self):
        return self.parents[0].total_devices

    def _get_member_devices(self):
        return self.parents[0].member_devices

    def _add_parent(self, parent):
        # pylint: disable=bad-super-call
        super(MDRaidArrayDevice, self)._add_parent(parent)

        if self.status and parent.format.exists:
            # we always probe since the device may not be set up when we want
            # information about it
            self._size = self.current_size

    def _remove_parent(self, parent):
        error_msg = self._validate_parent_removal(self.level, parent)
        if error_msg:
            raise errors.DeviceError(error_msg)

        # pylint: disable=bad-super-call
        super(MDRaidArrayDevice, self)._remove_parent(parent)

    def teardown(self, recursive=None):
        log_method_call(self, self.name, status=self.status,
                        controllable=self.controllable)
        # we don't really care about the return value of _pre_teardown here.
        # see comment just above md_deactivate call
        self._pre_teardown(recursive=recursive)

        # Since BIOS RAID sets (containers in mdraid terminology) never change
        # there is no need to stop them and later restart them. Not stopping
        # (and thus also not starting) them also works around bug 523334
        return



# availability.py
# Class for tracking availability of an application.
#
# Copyright (C) 2014-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc
import shutil

from six import add_metaclass

from .. import safe_dbus
from ..devicelibs.stratis import STRATIS_SERVICE, STRATIS_PATH

import gi
gi.require_version("BlockDev", "3.0")
gi.require_version("GLib", "2.0")
gi.require_version("Gio", "2.0")

from gi.repository import BlockDev as blockdev
from gi.repository import GLib, Gio

import logging
log = logging.getLogger("blivet")

CACHE_AVAILABILITY = True


class ExternalResource(object):

    """ An external resource. """

    def __init__(self, method, name):
        """ Initializes an instance of an external resource.

            :param method: A method object
            :type method: :class:`Method`
            :param str name: the name of the external resource
        """
        self._method = method
        self.name = name
        self._availability_errors = None

    def __str__(self):
        return self.name

    @property
    def availability_errors(self):
        """ Whether the resource has any availability errors.

            :returns: [] if the resource is available
            :rtype: list of str
        """
        _errors = list()

        # Prepare error cache and return value based on current caching setting.
        if CACHE_AVAILABILITY:
            _errors = self._availability_errors
        else:
            self._availability_errors = None

        # Check for errors if necessary.
        if self._availability_errors is None:
            _errors = self._method.availability_errors(self)

        # Update error cache if necessary.
        if CACHE_AVAILABILITY and self._availability_errors is None:
            self._availability_errors = _errors[:]

        return _errors

    @property
    def available(self):
        """ Whether the resource is available.

            :returns: True if the resource is available
            :rtype: bool
        """
        return self.availability_errors == []


@add_metaclass(abc.ABCMeta)
class Method(object):

    """ Method for determining if external resource is available."""

    @abc.abstractmethod
    def availability_errors(self, resource):
        """ Returns [] if the resource is available.

            :param resource: any external resource
            :type resource: :class:`ExternalResource`

            :returns: [] if the external resource is available
            :rtype: list of str
        """
        raise NotImplementedError()


class Path(Method):

    """ Methods for when application is found in  PATH. """

    def availability_errors(self, resource):
        """ Returns [] if the name of the application is in the path.

            :param resource: any application
            :type resource: :class:`ExternalResource`

            :returns: [] if the name of the application is in the path
            :rtype: list of str
        """
        if not shutil.which(resource.name):
            return ["application %s is not in $PATH" % resource.name]
        else:
            return []


Path = Path()


class AppVersionInfo(object):

    def __init__(self, app_name, required_version, version_opt, version_regex):
        """ Initializer.

            :param str app_name: the name of the application
            :param required_version: the required version for this application
            :param version_opt: command line option to print version of this application
            :param version_regex: regular expression to extract version from
                                  output of @version_opt
            :type required_version: :class:`distutils.LooseVersion` or NoneType
        """
        self.app_name = app_name
        self.required_version = required_version
        self.version_opt = version_opt
        self.version_regex = version_regex

    def __str__(self):
        return "%s-%s" % (self.app_name, self.required_version)


class VersionMethod(Method):

    """ Methods for checking the version of the external resource. """

    def __init__(self, version_info=None):
        """ Initializer.

            :param :class:`AppVersionInfo` version_info:
        """
        self.version_info = version_info
        self._availability_errors = None

    def availability_errors(self, resource):
        if self._availability_errors is not None and CACHE_AVAILABILITY:
            return self._availability_errors[:]

        self._availability_errors = Path.availability_errors(resource)

        if self.version_info.required_version is None:
            return self._availability_errors[:]

        try:
            ret = blockdev.utils.check_util_version(self.version_info.app_name,
                                                    self.version_info.required_version,
                                                    self.version_info.version_opt,
                                                    self.version_info.version_regex)
            if not ret:
                err = "installed version of %s is less than " \
                      "required version %s" % (self.version_info.app_name,
                                               self.version_info.required_version)
                self._availability_errors.append(err)
        except blockdev.UtilsError as e:
            err = "failed to get installed version of %s: %s" % (self.version_info.app_name, e)
            self._availability_errors.append(err)

        return self._availability_errors[:]


class BlockDevTechInfo(object):

    def __init__(self, plugin_name, check_fn, technologies):
        """ Initializer.

            :param str plugin_name: the name of the libblockdev plugin
            :param check_fn: function used to check for support availability
            :param technologies: list of required technologies
        """
        self.plugin_name = plugin_name
        self.check_fn = check_fn
        self.technologies = technologies

    def __str__(self):
        return "blockdev-%s" % self.plugin_name


class BlockDevMethod(Method):

    """ Methods for when application is actually a libblockdev plugin. """

    def __init__(self, tech_info):
        """ Initializer.

            :param :class:`AppVersionInfo` version_info:
        """
        self._tech_info = tech_info
        self._availability_errors = None

    def _check_technologies(self):
        errors = []
        for tech, mode in self._tech_info.technologies.items():
            try:
                self._tech_info.check_fn(tech, mode)
            except GLib.GError as e:
                errors.append("%s: %s" % (tech.value_name, e.message))
        return errors

    def availability_errors(self, resource):
        """ Returns [] if the plugin is loaded.

            :param resource: a libblockdev plugin
            :type resource: :class:`ExternalResource`

            :returns: [] if the name of the plugin is loaded
            :rtype: list of str
        """
        if self._tech_info.plugin_name not in blockdev.get_available_plugin_names():  # pylint: disable=no-value-for-parameter
            return ["libblockdev plugin %s not loaded" % self._tech_info.plugin_name]
        else:
            tech_missing = self._check_technologies()
            if tech_missing:
                return ["libblockdev plugin %s is loaded but some required "
                        "technologies are not available (%s)" % (self._tech_info.plugin_name, "; ".join(tech_missing))]
            else:
                return []


class DBusMethod(Method):

    """ Methods for when application is actually a DBus service. """

    def __init__(self, dbus_name, dbus_path):
        """ Initializer.

            :param :class:`AppVersionInfo` version_info:
        """
        self.dbus_name = dbus_name
        self.dbus_path = dbus_path
        self._availability_errors = None

    def availability_errors(self, resource):
        """ Returns [] if the service is available.

            :param resource: a DBus service
            :type resource: :class:`ExternalResource`

            :returns: [] if the name of the plugin is loaded
            :rtype: list of str
        """
        try:
            avail = blockdev.utils.dbus_service_available(None, Gio.BusType.SYSTEM, self.dbus_name, self.dbus_path)
            avail = safe_dbus.check_object_available(self.dbus_name, self.dbus_path)
        except safe_dbus.DBusCallError:
            return ["DBus service %s not available" % resource.name]
        else:
            if avail:
                return []
            else:
                return ["DBus service %s not available" % resource.name]


class _UnavailableMethod(Method):

    """ Method that indicates a resource is unavailable. """

    def __init__(self, error_msg=None):
        self.error_msg = error_msg or "always unavailable"

    def availability_errors(self, resource):
        return [self.error_msg]


UnavailableMethod = _UnavailableMethod()


class _AvailableMethod(Method):

    """ Method that indicates a resource is available. """

    def availability_errors(self, resource):
        return []


AvailableMethod = _AvailableMethod()


def application(name):
    """ Construct an external resource that is an application.

        This application will be available if its name can be found in $PATH.
    """
    return ExternalResource(Path, name)


def application_by_version(name, version_method):
    """ Construct an external resource that is an application.

        This application will be available if its name can be found in $PATH
        AND its version is at least the required version.

        :param :class:`VersionMethod` version_method: the version method
    """
    return ExternalResource(version_method, name)


def blockdev_plugin(name, blockdev_method):
    """ Construct an external resource that is a libblockdev plugin. """
    return ExternalResource(blockdev_method, name)


def dbus_service(name, dbus_method):
    """ Construct an external resource that is a DBus service. """
    return ExternalResource(dbus_method, name)


def unavailable_resource(name):
    """ Construct an external resource that is always unavailable. """
    return ExternalResource(UnavailableMethod, name)


def available_resource(name):
    """ Construct an external resource that is always available. """
    return ExternalResource(AvailableMethod, name)


# libblockdev btrfs plugin required technologies and modes
BLOCKDEV_BTRFS_ALL_MODES = (blockdev.BtrfsTechMode.CREATE |
                            blockdev.BtrfsTechMode.DELETE |
                            blockdev.BtrfsTechMode.MODIFY |
                            blockdev.BtrfsTechMode.QUERY)
BLOCKDEV_BTRFS = BlockDevTechInfo(plugin_name="btrfs",
                                  check_fn=blockdev.btrfs_is_tech_avail,
                                  technologies={blockdev.BtrfsTech.MULTI_DEV: BLOCKDEV_BTRFS_ALL_MODES,
                                                blockdev.BtrfsTech.SUBVOL: BLOCKDEV_BTRFS_ALL_MODES,
                                                blockdev.BtrfsTech.SNAPSHOT: BLOCKDEV_BTRFS_ALL_MODES})
BLOCKDEV_BTRFS_TECH = BlockDevMethod(BLOCKDEV_BTRFS)

# libblockdev crypto plugin required technologies and modes
BLOCKDEV_CRYPTO_ALL_MODES = (blockdev.CryptoTechMode.CREATE |
                             blockdev.CryptoTechMode.OPEN_CLOSE |
                             blockdev.CryptoTechMode.QUERY |
                             blockdev.CryptoTechMode.ADD_KEY |
                             blockdev.CryptoTechMode.REMOVE_KEY |
                             blockdev.CryptoTechMode.RESIZE)
BLOCKDEV_CRYPTO = BlockDevTechInfo(plugin_name="crypto",
                                   check_fn=blockdev.crypto_is_tech_avail,
                                   technologies={blockdev.CryptoTech.LUKS: BLOCKDEV_CRYPTO_ALL_MODES,
                                                 blockdev.CryptoTech.ESCROW: blockdev.CryptoTechMode.CREATE})
BLOCKDEV_CRYPTO_TECH = BlockDevMethod(BLOCKDEV_CRYPTO)

BLOCKDEV_CRYPTO_INTEGRITY = BlockDevTechInfo(plugin_name="crypto",
                                             check_fn=blockdev.crypto_is_tech_avail,
                                             technologies={blockdev.CryptoTech.INTEGRITY: (blockdev.CryptoTechMode.CREATE |
                                                                                           blockdev.CryptoTechMode.OPEN_CLOSE |
                                                                                           blockdev.CryptoTechMode.QUERY)})
BLOCKDEV_CRYPTO_TECH_INTEGRITY = BlockDevMethod(BLOCKDEV_CRYPTO_INTEGRITY)

# libblockdev dm plugin required technologies and modes
BLOCKDEV_DM_ALL_MODES = (blockdev.DMTechMode.CREATE_ACTIVATE |
                         blockdev.DMTechMode.REMOVE_DEACTIVATE |
                         blockdev.DMTechMode.QUERY)
BLOCKDEV_DM = BlockDevTechInfo(plugin_name="dm",
                               check_fn=blockdev.dm_is_tech_avail,
                               technologies={blockdev.DMTech.MAP: BLOCKDEV_DM_ALL_MODES})
BLOCKDEV_DM_TECH = BlockDevMethod(BLOCKDEV_DM)

# libblockdev loop plugin required technologies and modes
BLOCKDEV_LOOP_ALL_MODES = (blockdev.LoopTechMode.CREATE |
                           blockdev.LoopTechMode.CREATE |
                           blockdev.LoopTechMode.DESTROY |
                           blockdev.LoopTechMode.MODIFY |
                           blockdev.LoopTechMode.QUERY)
BLOCKDEV_LOOP = BlockDevTechInfo(plugin_name="loop",
                                 check_fn=blockdev.loop_is_tech_avail,
                                 technologies={blockdev.LoopTech.LOOP: BLOCKDEV_LOOP_ALL_MODES})
BLOCKDEV_LOOP_TECH = BlockDevMethod(BLOCKDEV_LOOP)

# libblockdev lvm plugin required technologies and modes
BLOCKDEV_LVM_ALL_MODES = (blockdev.LVMTechMode.CREATE |
                          blockdev.LVMTechMode.REMOVE |
                          blockdev.LVMTechMode.MODIFY |
                          blockdev.LVMTechMode.QUERY)
BLOCKDEV_LVM = BlockDevTechInfo(plugin_name="lvm",
                                check_fn=blockdev.lvm_is_tech_avail,
                                technologies={blockdev.LVMTech.BASIC: BLOCKDEV_LVM_ALL_MODES,
                                              blockdev.LVMTech.BASIC_SNAP: BLOCKDEV_LVM_ALL_MODES,
                                              blockdev.LVMTech.THIN: BLOCKDEV_LVM_ALL_MODES,
                                              blockdev.LVMTech.CACHE: BLOCKDEV_LVM_ALL_MODES,
                                              blockdev.LVMTech.CALCS: blockdev.LVMTechMode.QUERY,
                                              blockdev.LVMTech.THIN_CALCS: blockdev.LVMTechMode.QUERY,
                                              blockdev.LVMTech.CACHE_CALCS: blockdev.LVMTechMode.QUERY,
                                              blockdev.LVMTech.GLOB_CONF: (blockdev.LVMTechMode.QUERY |
                                                                           blockdev.LVMTechMode.MODIFY)})
BLOCKDEV_LVM_TECH = BlockDevMethod(BLOCKDEV_LVM)

if hasattr(blockdev.LVMTech, "VDO"):
    BLOCKDEV_LVM_VDO = BlockDevTechInfo(plugin_name="lvm",
                                        check_fn=blockdev.lvm_is_tech_avail,
                                        technologies={blockdev.LVMTech.VDO: (blockdev.LVMTechMode.CREATE |
                                                                             blockdev.LVMTechMode.REMOVE |
                                                                             blockdev.LVMTechMode.QUERY)})
    BLOCKDEV_LVM_TECH_VDO = BlockDevMethod(BLOCKDEV_LVM_VDO)
else:
    BLOCKDEV_LVM_TECH_VDO = _UnavailableMethod(error_msg="Installed version of libblockdev doesn't support LVM VDO technology")

if hasattr(blockdev.LVMTech, "SHARED"):
    BLOCKDEV_LVM_SHARED = BlockDevTechInfo(plugin_name="lvm",
                                           check_fn=blockdev.lvm_is_tech_avail,
                                           technologies={blockdev.LVMTech.SHARED: blockdev.LVMTechMode.MODIFY})  # pylint: disable=no-member
    BLOCKDEV_LVM_TECH_SHARED = BlockDevMethod(BLOCKDEV_LVM_SHARED)
else:
    BLOCKDEV_LVM_TECH_SHARED = _UnavailableMethod(error_msg="Installed version of libblockdev doesn't support shared LVM technology")

# libblockdev mdraid plugin required technologies and modes
BLOCKDEV_MD_ALL_MODES = (blockdev.MDTechMode.CREATE |
                         blockdev.MDTechMode.DELETE |
                         blockdev.MDTechMode.MODIFY |
                         blockdev.MDTechMode.QUERY)
BLOCKDEV_MD = BlockDevTechInfo(plugin_name="mdraid",
                               check_fn=blockdev.md_is_tech_avail,
                               technologies={blockdev.MDTech.MDRAID: BLOCKDEV_MD_ALL_MODES})
BLOCKDEV_MD_TECH = BlockDevMethod(BLOCKDEV_MD)

# libblockdev mpath plugin required technologies and modes
BLOCKDEV_MPATH_ALL_MODES = (blockdev.MpathTechMode.MODIFY |
                            blockdev.MpathTechMode.QUERY)
BLOCKDEV_MPATH = BlockDevTechInfo(plugin_name="mpath",
                                  check_fn=blockdev.mpath_is_tech_avail,
                                  technologies={blockdev.MpathTech.BASE: BLOCKDEV_MPATH_ALL_MODES})
BLOCKDEV_MPATH_TECH = BlockDevMethod(BLOCKDEV_MPATH)

# libblockdev swap plugin required technologies and modes
BLOCKDEV_SWAP_ALL_MODES = (blockdev.SwapTechMode.CREATE |
                           blockdev.SwapTechMode.ACTIVATE_DEACTIVATE |
                           blockdev.SwapTechMode.QUERY |
                           blockdev.SwapTechMode.SET_LABEL)
BLOCKDEV_SWAP = BlockDevTechInfo(plugin_name="swap",
                                 check_fn=blockdev.swap_is_tech_avail,
                                 technologies={blockdev.SwapTech.SWAP: BLOCKDEV_SWAP_ALL_MODES})
BLOCKDEV_SWAP_TECH = BlockDevMethod(BLOCKDEV_SWAP)

# libblockdev plugins
# we can't just check if the plugin is loaded, we also need to make sure
# that all technologies required by us our supported (some may be missing
# due to missing dependencies)
BLOCKDEV_BTRFS_PLUGIN = blockdev_plugin("libblockdev btrfs plugin", BLOCKDEV_BTRFS_TECH)
BLOCKDEV_CRYPTO_PLUGIN = blockdev_plugin("libblockdev crypto plugin", BLOCKDEV_CRYPTO_TECH)
BLOCKDEV_CRYPTO_PLUGIN_INTEGRITY = blockdev_plugin("libblockdev crypto plugin (integrity technology)",
                                                   BLOCKDEV_CRYPTO_TECH_INTEGRITY)
BLOCKDEV_DM_PLUGIN = blockdev_plugin("libblockdev dm plugin", BLOCKDEV_DM_TECH)
BLOCKDEV_LOOP_PLUGIN = blockdev_plugin("libblockdev loop plugin", BLOCKDEV_LOOP_TECH)
BLOCKDEV_LVM_PLUGIN = blockdev_plugin("libblockdev lvm plugin", BLOCKDEV_LVM_TECH)
BLOCKDEV_LVM_PLUGIN_VDO = blockdev_plugin("libblockdev lvm plugin (vdo technology)", BLOCKDEV_LVM_TECH_VDO)
BLOCKDEV_LVM_PLUGIN_SHARED = blockdev_plugin("libblockdev lvm plugin (shared LVM technology)", BLOCKDEV_LVM_TECH_SHARED)
BLOCKDEV_MDRAID_PLUGIN = blockdev_plugin("libblockdev mdraid plugin", BLOCKDEV_MD_TECH)
BLOCKDEV_MPATH_PLUGIN = blockdev_plugin("libblockdev mpath plugin", BLOCKDEV_MPATH_TECH)
BLOCKDEV_SWAP_PLUGIN = blockdev_plugin("libblockdev swap plugin", BLOCKDEV_SWAP_TECH)

# applications with versions
# we need e2fsprogs newer than 1.41 and we are checking the version by running
# the "e2fsck" tool and parsing its ouput for version number
E2FSPROGS_INFO = AppVersionInfo(app_name="e2fsck",
                                required_version="1.41.0",
                                version_opt="-V",
                                version_regex=r"e2fsck ([0-9+\.]+)[\-rc0-9+]* .*")
E2FSPROGS_VERSION = VersionMethod(E2FSPROGS_INFO)


# new version of dosftools changed behaviour of many tools
DOSFSTOOLS_INFO = AppVersionInfo(app_name="mkdosfs",
                                 required_version="4.2",
                                 version_opt="--help",
                                 version_regex=r"mkfs\.fat ([0-9+\.]+) .*")
DOSFSTOOLS_VERSION = VersionMethod(DOSFSTOOLS_INFO)

# applications
DEBUGREISERFS_APP = application("debugreiserfs")
DF_APP = application("df")
DOSFSCK_APP = application("dosfsck")
DOSFSLABEL_APP = application("dosfslabel")
DUMPE2FS_APP = application_by_version("dumpe2fs", E2FSPROGS_VERSION)
E2FSCK_APP = application_by_version("e2fsck", E2FSPROGS_VERSION)
E2LABEL_APP = application_by_version("e2label", E2FSPROGS_VERSION)
FSCK_HFSPLUS_APP = application("fsck.hfsplus")
HFORMAT_APP = application("hformat")
JFSTUNE_APP = application("jfs_tune")
KPARTX_APP = application("kpartx")
LVMDEVICES = application("lvmdevices")
MKDOSFS_APP = application("mkdosfs")
MKDOSFS_NEW_APP = application_by_version("mkdosfs", DOSFSTOOLS_VERSION)
MKE2FS_APP = application_by_version("mke2fs", E2FSPROGS_VERSION)
MKFS_BTRFS_APP = application("mkfs.btrfs")
MKFS_GFS2_APP = application("mkfs.gfs2")
MKFS_HFSPLUS_APP = application("mkfs.hfsplus")
MKFS_JFS_APP = application("mkfs.jfs")
MKFS_XFS_APP = application("mkfs.xfs")
MKNTFS_APP = application("mkntfs")
MKREISERFS_APP = application("mkreiserfs")
MLABEL_APP = application("mlabel")
MULTIPATH_APP = application("multipath")
NTFSINFO_APP = application("ntfsinfo")
NTFSLABEL_APP = application("ntfslabel")
NTFSRESIZE_APP = application("ntfsresize")
REISERFSTUNE_APP = application("reiserfstune")
RESIZE2FS_APP = application_by_version("resize2fs", E2FSPROGS_VERSION)
TUNE2FS_APP = application_by_version("tune2fs", E2FSPROGS_VERSION)
XFSADMIN_APP = application("xfs_admin")
XFSDB_APP = application("xfs_db")
XFSFREEZE_APP = application("xfs_freeze")
XFSRESIZE_APP = application("xfs_growfs")
XFSREPAIR_APP = application("xfs_repair")

FSCK_F2FS_APP = application("fsck.f2fs")
MKFS_F2FS_APP = application("mkfs.f2fs")

MOUNT_APP = application("mount")

STRATISPREDICTUSAGE_APP = application("stratis-predict-usage")

STRATIS_SERVICE_METHOD = DBusMethod(STRATIS_SERVICE, STRATIS_PATH)
STRATIS_DBUS = dbus_service("stratis", STRATIS_SERVICE_METHOD)



# devices/network.py
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from .lib import Tags


class NetworkStorageDevice(object):

    """ Virtual base class for network backed storage devices """

    def __init__(self, host_address=None, nic=None):
        """ Note this class is only to be used as a baseclass and then only with
            multiple inheritance. The only correct use is:
            class MyStorageDevice(StorageDevice, NetworkStorageDevice):

            The sole purpose of this class is to:
            1) Be able to check if a StorageDevice is network backed
               (using isinstance).
            2) To be able to get the host address of the host (server) backing
               the storage *or* the NIC through which the storage is connected

            :keyword host_address: host address of the backing server
            :type host_address: str
            :keyword nic: NIC to which the block device is bound
            :type nic: str
        """
        self.host_address = host_address
        self.nic = nic

        self.tags.add(Tags.remote)  # pylint: disable=no-member



# populator/helpers/devicepopulator.py
# Base class for device-type-specific helpers for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from .populatorhelper import PopulatorHelper
from ... import udev


# pylint: disable=abstract-method
class DevicePopulator(PopulatorHelper):
    """ Populator helper base class for devices.

        Subclasses must define a match method and, if they want to instantiate
        a device, a run method.
    """
    @classmethod
    def match(cls, data):
        return False

    def _handle_rename(self):
        name = udev.device_get_name(self.data)
        if self.device.name != name:
            self.device.name = name
        # TODO: update name registry -- better yet, generate the name list on demand

    def _handle_resize(self):
        old_size = self.device.current_size
        self.device.update_size()
        if old_size != self.device.current_size:
            self._devicetree.cancel_disk_actions(self.device.disks)

    def update(self):
        self._handle_rename()
        self._handle_resize()



# __init__.py
#
# Copyright (C) 2009, 2010, 2011, 2012, 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#                    Vratislav Podzimek <vpodzime@redhat.com>
#

__version__ = '3.8.2'

import sys
import importlib
import warnings
import syslog

from . import util, arch

import logging
log = logging.getLogger("blivet")
program_log = logging.getLogger("program")
testdata_log = logging.getLogger("testdata")

# Tell the warnings module not to ignore DeprecationWarning, which it does by
# default since python-2.7.
warnings.simplefilter('module', DeprecationWarning)

# Enable logging of python warnings.
logging.captureWarnings(True)

# "translation" between syslog log levels (used by libblockdev) and python log levels
LOG_LEVELS = {syslog.LOG_EMERG: logging.CRITICAL, syslog.LOG_ALERT: logging.CRITICAL,
              syslog.LOG_CRIT: logging.CRITICAL, syslog.LOG_ERR: logging.ERROR,
              syslog.LOG_WARNING: logging.WARNING, syslog.LOG_NOTICE: logging.INFO,
              syslog.LOG_INFO: logging.INFO, syslog.LOG_DEBUG: logging.DEBUG}


def log_bd_message(level, msg):
    # only log <= info for libblockdev, debug contains debug messages
    # from cryptsetup and we don't want to put these into program.log
    if level <= syslog.LOG_INFO and level in LOG_LEVELS.keys():
        program_log.log(LOG_LEVELS[level], msg)


import gi
gi.require_version("GLib", "2.0")
gi.require_version("BlockDev", "3.0")

# initialize the libblockdev library
from gi.repository import GLib
from gi.repository import BlockDev as blockdev
if arch.is_s390():
    _REQUESTED_PLUGIN_NAMES = set(("lvm", "btrfs", "swap", "crypto", "loop", "mdraid", "mpath", "dm", "s390", "nvdimm", "nvme"))
else:
    _REQUESTED_PLUGIN_NAMES = set(("lvm", "btrfs", "swap", "crypto", "loop", "mdraid", "mpath", "dm", "nvdimm", "nvme"))

_requested_plugins = blockdev.plugin_specs_from_names(_REQUESTED_PLUGIN_NAMES)
try:
    succ_, avail_plugs = blockdev.try_reinit(require_plugins=_requested_plugins, reload=False, log_func=log_bd_message)
except GLib.GError as err:
    raise RuntimeError("Failed to initialize the libblockdev library: %s" % err)
else:
    avail_plugs = set(avail_plugs)

missing_plugs = _REQUESTED_PLUGIN_NAMES - avail_plugs
for p in missing_plugs:
    log.info("Failed to load plugin %s", p)


class _LazyImportObject(object):

    """
    A simple class that uses sys.modules and importlib to implement a
    lazy-imported object. Once it is called (or instantiated) or an attribute of
    it is requested, the real object is imported and an appropriate method is
    called on it with all the passed arguments.

    """

    def __init__(self, name, real_mod):
        """
        Create a new instance of a lazy-imported object.

        :param str name: name of the real object/class
        :param str real_mod: the real module the real object lives in

        """

        self._name = name
        self._real_mod = real_mod

    def __call__(self, *args, **kwargs):
        mod = importlib.import_module(__package__ + "." + self._real_mod)
        val = getattr(mod, self._name)
        sys.modules["%s.%s" % (__package__, self._name)] = val
        return val(*args, **kwargs)

    def __getattr__(self, attr):
        mod = importlib.import_module(__package__ + "." + self._real_mod)
        val = getattr(mod, self._name)
        sys.modules["%s.%s" % (__package__, self._name)] = val
        return getattr(val, attr)

    def __dir__(self):
        mod = importlib.import_module(__package__ + "." + self._real_mod)
        val = getattr(mod, self._name)
        sys.modules["%s.%s" % (__package__, self._name)] = val
        return dir(val)


# this way things like 'from blivet import Blivet' work without an overhead of
# importing of everything the Blivet class needs whenever anything from the
# 'blivet' package is imported (e.g. the 'arch' module)
Blivet = _LazyImportObject("Blivet", "blivet")



# devices/storage.py
# Base class for block device classes.
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import os
import copy
import pyudev

from ..callbacks import callbacks
from .. import errors
from .. import util
from ..flags import flags
from ..storage_log import log_method_call
from .. import udev
from ..formats import get_format, DeviceFormat
from ..size import Size

import logging
log = logging.getLogger("blivet")

from .device import Device
from .network import NetworkStorageDevice
from .lib import LINUX_SECTOR_SIZE
from ..devicelibs.crypto import LUKS_METADATA_SIZE


class StorageDevice(Device):

    """ A generic storage device.

        A fully qualified path to the device node can be obtained via the
        path attribute, although it is not guaranteed to be useful, or
        even present, unless the StorageDevice's setup method has been
        run.
    """
    _resizable = False
    """Whether this type of device is inherently resizable."""
    _renamable = False

    _type = "blivet"
    _dev_dir = "/dev"
    _format_immutable = False
    _partitionable = False
    _is_disk = False
    _encrypted = False

    config_actions_map = {"name": "_rename"}

    def __init__(self, name, fmt=None, uuid=None,
                 size=None, major=None, minor=None,
                 sysfs_path='', parents=None, exists=False, serial=None,
                 vendor="", model="", bus=""):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword uuid: universally unique identifier (device -- not fs)
            :type uuid: str
            :keyword sysfs_path: sysfs device path
            :type sysfs_path: str
            :keyword major: the device major
            :type major: int
            :keyword minor: the device minor
            :type minor: int
            :keyword serial: the ID_SERIAL_SHORT for this device
            :type serial: str
            :keyword vendor: the manufacturer of this Device
            :type vendor: str
            :keyword model: manufacturer's device model string
            :type model: str
            :keyword bus: the interconnect this device uses
            :type bus: str

        """
        # allow specification of individual parents
        if isinstance(parents, Device):
            parents = [parents]

        self.exists = exists
        self.uuid = uuid

        # Set sysfs path before super call as MDRaidArrayDevice._add_parent()
        # reads it via status().
        self.sysfs_path = sysfs_path

        self._format = get_format(None)

        # For non-existent devices, make sure the initial size is enough for
        # the format's metadata. This is mostly relevant for growable
        # partitions and lvs with thoughtless initial sizes.
        if not self.exists and fmt and fmt.min_size:
            min_size = max(util.numeric_type(size), fmt.min_size)
            if min_size > util.numeric_type(size):
                log.info("%s: using size %s instead of %s to accommodate "
                         "format minimum size", name, min_size, size)
                size = min_size

        # The size will be overridden by a call to update_size at the end of this
        # method for existing and active devices.
        self._size = Size(util.numeric_type(size))
        self._target_size = self._size
        self._current_size = self._size if self.exists else Size(0)
        self.major = util.numeric_type(major)
        self.minor = util.numeric_type(minor)
        self._serial = serial
        self._vendor = vendor or ""
        self._model = model or ""
        self.bus = bus

        self._logical_block_size = None
        self._physical_block_size = None

        self._readonly = False
        self._protected = False
        self._controllable = True

        # Copy only the validity check from Device._set_name() so we don't try
        # to check a bunch of inappropriate state properties during
        # __init__ in subclasses
        # Has to be here because Device does not have exists attribute
        if not self.exists and not self.is_name_valid(name):
            raise ValueError("%s is not a valid name for this device" % name)
        super(StorageDevice, self).__init__(name, parents=parents)

        self.format = fmt
        self.original_format = copy.deepcopy(self.format)
        self.fstab_comment = ""

        self.device_links = []

        if self.exists:
            if self.status:
                self.update_sysfs_path()
                self.update_size()

    def __str__(self):
        exist = "existing"
        if not self.exists:
            exist = "non-existent"
        s = "%s %s %s" % (exist, self.size, super(StorageDevice, self).__str__())
        if self.format.type:
            s += " with %s" % self.format

        return s

    @property
    def packages(self):
        packages = super(StorageDevice, self).packages
        packages.extend(p for p in self.format.packages if p not in packages)
        return packages

    @property
    def disks(self):
        """ A list of all disks this device depends on, including itself. """
        _disks = []
        for parent in self.parents:
            _disks.extend(d for d in parent.disks if d not in _disks)

        if self.is_disk and not self.format.hidden:
            _disks.append(self)

        return _disks

    @property
    def encrypted(self):
        """ True if this device, or any it requires, is encrypted. """
        return self._encrypted or any(p.encrypted for p in self.parents)

    @property
    def raw_device(self):
        """ The device itself, or when encrypted, the backing device. """
        return self

    @property
    def sector_size(self):
        """ Logical sector (block) size of this device """
        return self.logical_block_size

    @property
    def logical_block_size(self):
        """ Logical sector (block) size of this device """
        if not self.exists:
            if self.parents:
                return self.parents[0].logical_block_size
            else:
                return LINUX_SECTOR_SIZE

        if self._logical_block_size:
            return self._logical_block_size

        block_size = util.get_sysfs_attr(self.sysfs_path, "queue/logical_block_size")
        if block_size:
            self._logical_block_size = int(block_size)
            return self._logical_block_size
        else:
            return LINUX_SECTOR_SIZE

    @property
    def physical_block_size(self):
        """ Physical sector (block) size of this device """
        if not self.exists:
            if self.parents:
                return self.parents[0].physical_block_size
            else:
                return LINUX_SECTOR_SIZE

        if self._physical_block_size:
            return self._physical_block_size

        block_size = util.get_sysfs_attr(self.sysfs_path, "queue/physical_block_size")
        if block_size:
            self._physical_block_size = int(block_size)
            return self._physical_block_size
        else:
            return LINUX_SECTOR_SIZE

    @property
    def controllable(self):
        return self._controllable and not flags.testing and not self.unavailable_type_dependencies()

    @controllable.setter
    def controllable(self, value):
        self._controllable = value

    def _set_name(self, value):
        """Set the device's name.

        :param value: the new device name
        :raises errors.DeviceError: if the device exists
        """

        if value == self._name:
            return

        super(StorageDevice, self)._set_name(value)

        # update our format's path
        # First, check that self._format has been defined in case this is
        # running early in the constructor.
        if hasattr(self, "_format") and self.format.device:
            self.format.device = self.path

    def align_target_size(self, newsize):
        """ Return a proposed target size adjusted for device specifics.

            :param :class:`~.Size` newsize: the proposed/unaligned target size
            :returns: newsize modified to yield an aligned device
            :rtype: :class:`~.Size`
        """

        return newsize

    def _get_target_size(self):
        return self._target_size

    def _set_target_size(self, newsize):
        if not isinstance(newsize, Size):
            raise ValueError("new size must of type Size")

        if self.max_size and newsize > self.max_size:
            log.error("requested size %s is larger than maximum %s",
                      newsize, self.max_size)
            raise ValueError("size is larger than the maximum for this device")
        elif self.min_size and newsize < self.min_size:
            log.error("requested size %s is smaller than minimum %s",
                      newsize, self.min_size)
            raise ValueError("size is smaller than the minimum for this device")

        if self.align_target_size(newsize) != newsize:
            raise ValueError("new size would violate alignment requirements")

        self._target_size = newsize

    target_size = property(lambda s: s._get_target_size(),
                           lambda s, v: s._set_target_size(v),
                           doc="Target size of this device")

    def __repr__(self):
        s = Device.__repr__(self)
        s += ("  uuid = %(uuid)s  size = %(size)s\n"
              "  format = %(format)s\n"
              "  major = %(major)s  minor = %(minor)s  exists = %(exists)s"
              "  protected = %(protected)s\n"
              "  sysfs path = %(sysfs)s\n"
              "  target size = %(target_size)s  path = %(path)s\n"
              "  format args = %(format_args)s  original_format = %(orig_fmt)s" %
              {"uuid": self.uuid, "format": self.format, "size": self.size,
               "major": self.major, "minor": self.minor, "exists": self.exists,
               "sysfs": self.sysfs_path,
               "target_size": self.target_size, "path": self.path,
               "protected": self.protected,
               "format_args": self.format_args, "orig_fmt": self.original_format.type})
        return s

    @property
    def dict(self):
        d = super(StorageDevice, self).dict
        d.update({"uuid": self.uuid, "size": self.size,
                  "format": self.format.dict, "removable": self.removable,
                  "major": self.major, "minor": self.minor,
                  "exists": self.exists, "sysfs": self.sysfs_path,
                  "target_size": self.target_size, "path": self.path})
        return d

    @property
    def path(self):
        """ Device node representing this device. """
        return "%s/%s" % (self._dev_dir, self.name)

    def update_sysfs_path(self):
        """ Update this device's sysfs path. """
        # We're using os.path.exists as a stand-in for status. We can't use
        # the status property directly because MDRaidArrayDevice.status calls
        # this method.
        log_method_call(self, self.name, status=os.path.exists(self.path))
        if not self.exists:
            raise errors.DeviceError("device has not been created")

        try:
            udev_device = pyudev.Devices.from_device_file(udev.global_udev,
                                                          self.path)

        # from_device_file() does not process exceptions but just propagates
        # any errors that are raised.
        except (pyudev.DeviceNotFoundError, EnvironmentError, ValueError, OSError) as e:
            log.error("failed to update sysfs path for %s: %s", self.name, e)
            self.sysfs_path = ''
        else:
            self.sysfs_path = udev_device.sys_path
            log.debug("%s sysfs_path set to %s", self.name, self.sysfs_path)

    @property
    def format_args(self):
        """ Device-specific arguments to format creation program. """
        return []

    @property
    def resizable(self):
        """ Can this device be resized? """
        return (self._resizable and self.exists and
                (self.format.resizable or not self.format.exists))

    @property
    def fstab_spec(self):
        spec = self.path
        if self.format and self.format.uuid:
            spec = "UUID=%s" % self.format.uuid
        return spec

    def resize(self):
        """ Resize a device to self.target_size.

            This method should only be invoked via the
            ActionResizeDevice.execute method. All the pre-conditions
            enforced by ActionResizeDevice.__init__ are assumed to hold.

            Returns nothing.
        """
        if self._resizable:
            raise NotImplementedError("method not implemented for device type %s" % self.type)
        else:
            raise errors.DeviceError("device type %s is not resizable" % self.type)

    def _rename(self, old_name, new_name, dry_run=False):  # pylint: disable=unused-argument
        """ Rename this device.

            :param old_name: current name of this device
            :type old_name: str
            :param new_name: new name for this device
            :type new_name: str
            :param dry_run: whether to only run checks and not perform the rename
            :type dry_run: bool

            Note: This method should only be invoked via the
                  ActionConfigureDevice.execute method. All the pre-conditions
                  enforced by ActionConfigureDevice.__init__ are assumed to hold.
                  self.name value is chaged by the ActionConfigureDevice.apply method.

                  Caller must make sure the name is unique, this method cannot check
                  the name uniqueness.
        """
        if not self._renamable:
            raise errors.DeviceError("device type %s is not renamable" % self.type)
        else:
            raise NotImplementedError("method not implemented for device type %s" % self.type)

    @property
    def readonly(self):
        # A device is read-only if it or any parent device is read-only
        return self._readonly or any(p.readonly for p in self.parents)

    @readonly.setter
    def readonly(self, value):
        self._readonly = value

    @property
    def protected(self):
        return self.readonly or self._protected

    @protected.setter
    def protected(self, value):
        self._protected = value

    #
    # setup
    #
    def _pre_setup(self, orig=False):
        """ Preparation and pre-condition checking for device setup.

            Return True if setup should proceed or False if not.
        """
        if not self.exists:
            raise errors.DeviceError("device has not been created")

        if self.status or not self.controllable:
            return False

        self.setup_parents(orig=orig)
        return True

    def _setup(self, orig=False):
        """ Perform device-specific setup operations. """

    def setup(self, orig=False):
        """ Open, or set up, a device. """
        log_method_call(self, self.name, orig=orig, status=self.status,
                        controllable=self.controllable)
        if not self._pre_setup(orig=orig):
            return

        self._setup(orig=orig)
        self._post_setup()

    def _post_setup(self):
        """ Perform post-setup operations. """
        udev.settle()
        self.update_sysfs_path()
        # the device may not be set up when we want information about it
        if self._size == Size(0):
            self.update_size()

    #
    # teardown
    #
    def _pre_teardown(self, recursive=None):
        """ Preparation and pre-condition checking for device teardown.

            Return True if teardown should proceed or False if not.
        """
        if not self.exists and not recursive:
            raise errors.DeviceError("device has not been created")

        if not self.status or not self.controllable or self.protected:
            return False

        if self.original_format.exists:
            self.original_format.teardown()
        if self.format.exists:
            self.format.teardown()
        return True

    def _teardown(self, recursive=None):
        """ Perform device-specific teardown operations. """

    def teardown(self, recursive=None):
        """ Close, or tear down, a device. """
        log_method_call(self, self.name, status=self.status,
                        controllable=self.controllable)
        if not self._pre_teardown(recursive=recursive):
            if recursive:
                self.teardown_parents(recursive=recursive)
            return

        self._teardown(recursive=recursive)
        self._post_teardown(recursive=recursive)

    def _post_teardown(self, recursive=None):
        """ Perform post-teardown operations. """
        if recursive:
            self.teardown_parents(recursive=recursive)

    #
    # create
    #
    def _pre_create(self):
        """ Preparation and pre-condition checking for device creation. """
        if self.exists:
            raise errors.DeviceError("device has already been created")

        self.setup_parents()

    def _create(self):
        """ Perform device-specific create operations. """

    def create(self):
        """ Create the device. """
        log_method_call(self, self.name, status=self.status)
        self._pre_create()
        self._create()
        self._post_create()

    def _post_create(self):
        """ Perform post-create operations. """
        self.exists = True
        self.setup()
        self.update_sysfs_path()
        udev.settle()

        # make sure that target_size is updated to reflect the actual size
        self.update_size()

        self._update_netdev_mount_option()

    #
    # destroy
    #
    def _pre_destroy(self):
        """ Preparation and precondition checking for device destruction. """
        if not self.exists:
            raise errors.DeviceError("device has not been created")

        if not self.isleaf:
            raise errors.DeviceError("Cannot destroy non-leaf device %s" % self.name)

        self.teardown()

    def _destroy(self):
        """ Perform device-specific destruction operations. """

    def destroy(self):
        """ Destroy the device. """
        log_method_call(self, self.name, status=self.status)
        self._pre_destroy()
        self._destroy()
        self._post_destroy()

    def _post_destroy(self):
        """ Perform post-destruction operations. """
        self.exists = False

    #
    # parents' modifications/notifications
    #
    def setup_parents(self, orig=False):
        """ Run setup method of all parent devices. """
        log_method_call(self, name=self.name, orig=orig)
        for parent in self.parents:
            parent.setup(orig=orig)
            if orig:
                _format = parent.original_format
            else:
                _format = parent.format

            # set up the formatting, if present
            if _format.type and _format.exists:
                _format.setup()

    # pylint: disable=unused-argument
    def remove_hook(self, modparent=True):
        """ Perform actions related to removing a device from the devicetree.

            :keyword bool modparent: whether to account for removal in parents

            Parents' list of child devices is updated regardless of modparent.
            The intended use of modparent is to prevent doing things like
            removing a parted.Partition from the disk that contains it as part
            of msdos extended partition management. In general, you should not
            override the default value of modparent in new code.
        """
        for parent in self.parents:
            parent.remove_child(self)

    def add_hook(self, new=True):
        """ Perform actions related to adding a device to the devicetree.

            :keyword bool new: whether this device is new to the devicetree

            The only intended use case for new=False is when unhiding a device
            from the devicetree. Additional operations are performed when new is
            False that are normally performed as part of the device constructor.
        """
        if not new:
            for p in self.parents:
                p.add_child(self)

    #
    # size manipulations
    #
    def _get_size(self):
        """ Get the device's size, accounting for pending changes. """
        size = self._size
        if self.exists and self.resizable and self.target_size != Size(0):
            size = self.target_size

        return size

    def _set_size(self, newsize):
        """ Set the device's size to a new value.

            This is not adequate to set up a resize as it does not set a new
            target size for the device.
        """
        if not isinstance(newsize, Size):
            raise ValueError("new size must of type Size")

        # There's no point in checking limits here for existing devices since
        # the only way to change their size is by setting target size. Any call
        # to this setter for an existing device should be to reflect existing
        # state.
        if not self.exists:
            max_size = self.format.max_size
            min_size = self.format.min_size
            if max_size and newsize > max_size:
                raise errors.DeviceError("device cannot be larger than %s" %
                                         max_size, self.name)
            elif min_size and newsize < min_size:
                raise errors.DeviceError("device cannot be smaller than %s" %
                                         min_size, self.name)

        self._size = newsize

    size = property(lambda x: x._get_size(),
                    lambda x, y: x._set_size(y),
                    doc="The device's size, accounting for pending changes")

    def read_current_size(self):
        log_method_call(self, exists=self.exists, path=self.path,
                        sysfs_path=self.sysfs_path)
        size = Size(0)
        if self.exists and os.path.exists(self.path) and \
           os.path.isdir(self.sysfs_path):
            blocks = int(util.get_sysfs_attr(self.sysfs_path, "size") or '0')
            size = Size(blocks * LINUX_SECTOR_SIZE)

        return size

    @property
    def current_size(self):
        """ The device's actual size, generally the size discovered by using
            system tools. May use a cached value if the information is
            currently unavailable.

            If the device does not exist, then the actual size is 0.
        """
        if self._current_size == Size(0):
            self._current_size = self.read_current_size()
        return self._current_size

    def update_size(self, newsize=None):
        """ Update size, current_size, and target_size to actual size.

            :keyword :class:`~.size.Size` newsize: new size for device

            .. note::

                Most callers will not pass a new size. It is for special cases
                like outside resize of inactive LVs, which precludes updating
                the size from /sys.
        """
        if newsize is None:
            self._current_size = Size(0)
        elif isinstance(newsize, Size):
            self._current_size = newsize
        else:
            raise ValueError("new size must be an instance of class Size")

        new_size = self.current_size
        self._size = new_size
        self._target_size = new_size  # bypass setter checks
        log.debug("updated %s size to %s (%s)", self.name, self.size, new_size)

    @property
    def min_size(self):
        """ The minimum size this device can be. """
        if self.format.type == "luks" and self.children:
            if self.resizable:
                min_size = self.children[0].min_size + LUKS_METADATA_SIZE
            else:
                min_size = self.current_size
        else:
            min_size = self.align_target_size(self.format.min_size) if self.resizable else self.current_size

        return min_size

    @property
    def max_size(self):
        """ The maximum size this device can be. """
        return self.align_target_size(self.format.max_size) if self.resizable else self.current_size

    @property
    def growable(self):
        """ True if this device or its component devices are growable. """
        return getattr(self, "req_grow", False) or any(p.growable for p in self.parents)

    def check_size(self):
        """ Check to make sure the size of the device is allowed by the
            format used.

            Returns:
            0  - ok
            1  - Too large
            -1 - Too small
        """
        if self.format.max_size and self.size > self.format.max_size:
            return 1
        elif self.format.min_size and self.size < self.format.min_size:
            return -1
        return 0

    #
    # status
    #
    @property
    def media_present(self):
        """ True if this device contains usable media. """
        return True

    @property
    def status(self):
        """ This device's status.

            For now, this should return a boolean:
                True    the device is open and ready for use
                False   the device is not open
        """
        if not self.exists:
            return False
        return os.access(self.path, os.W_OK)

    #
    # format manipulations
    #
    def _set_format(self, fmt):
        """ Set the Device's format.

            :param fmt: the new format or None
            :type fmt: :class:`~.formats.DeviceFormat` or NoneType

            A value of None will effectively mark the device as unformatted,
            but this is accomplished by setting it to an instance of the base
            :class:`~.formats.DeviceFormat` class.

            .. note::
                :attr:`format` should always be an instance of
                :class:`~.formats.DeviceFormat`. To ensure this continues to be
                the case, all subclasses that define their own :attr:`format`
                setter should call :meth:`StorageDevice._set_format` from their
                setter.

        """
        if not fmt:
            fmt = get_format(None, device=self.path, exists=self.exists)

        if not isinstance(fmt, DeviceFormat):
            raise ValueError("format must be a DeviceFormat instance")

        log_method_call(self, self.name, type=fmt.type,
                        current=getattr(self._format, "type", None))

        # check device size against format limits
        if not fmt.exists:
            if fmt.max_size and fmt.max_size < self.size:
                raise errors.DeviceError("device is too large for new format")
            elif fmt.min_size and fmt.min_size > self.size:
                if self.growable:
                    log.info("%s: using size %s instead of %s to accommodate "
                             "format minimum size", self.name, fmt.min_size, self.size)
                    self.size = fmt.min_size
                else:
                    raise errors.DeviceError("device is too small for new format")

        if self._format != fmt:
            callbacks.format_removed(device=self, fmt=self._format)
            self._format = fmt
            self._format.device = self.path
            self._update_netdev_mount_option()
            callbacks.format_added(device=self, fmt=self._format)

    def _update_netdev_mount_option(self):
        """ Fix mount options to include or exclude _netdev as appropriate. """
        if not hasattr(self._format, "mountpoint"):
            return

        netdev_option = "_netdev"
        option_list = self._format.options.split(",")
        user_options = self._format._user_mountopts.split(",")
        is_netdev = any(isinstance(a, NetworkStorageDevice)
                        for a in self.ancestors)
        has_netdev_option = netdev_option in option_list
        if not is_netdev and has_netdev_option and netdev_option not in user_options:
            option_list.remove(netdev_option)
            self._format.options = ",".join(option_list)
        elif is_netdev and not has_netdev_option:
            option_list.append(netdev_option)
            self._format.options = ",".join(option_list)

    def _get_format(self):
        """ Get the device's format instance.

            :returns: this device's format instance
            :rtype: :class:`~.formats.DeviceFormat`

            .. note::
                :attr:`format` should always be an instance of
                :class:`~.formats.DeviceFormat`. Under no circumstances should
                a programmer directly set :attr:`_format` to any other type.

        """
        return self._format

    format = property(lambda d: d._get_format(),
                      lambda d, f: d._set_format(f),
                      doc="The device's formatting.")

    def pre_commit_fixup(self, current_fmt=False):
        """ Do any necessary pre-commit fixups."""

    @property
    def format_immutable(self):
        """ Is it possible to execute format actions on this device? """
        return self._format_immutable or self.protected

    #
    # misc properties
    #
    @property
    def removable(self):
        devpath = os.path.normpath(self.sysfs_path)
        remfile = os.path.normpath("%s/removable" % devpath)
        return (self.sysfs_path and os.path.exists(devpath) and
                os.access(remfile, os.R_OK) and
                open(remfile).readline().strip() == "1")

    @property
    def direct(self):
        """ Is this device directly accessible? """
        return self.isleaf

    @property
    def is_disk(self):
        return self._is_disk

    @property
    def is_empty(self):
        if not self.partitioned:
            return self.format.type is None and len(self.children) == 0

        return all(p.type == "partition" and p.is_magic for p in self.children)

    @property
    def partitionable(self):
        return self._partitionable

    @property
    def partitioned(self):
        return self.format.type == "disklabel" and self.partitionable

    @property
    def serial(self):
        return self._serial

    @property
    def model(self):
        return self._model

    @property
    def vendor(self):
        return self._vendor

    def populate_ksdata(self, data):
        # the common pieces are basically the formatting
        self.format.populate_ksdata(data)

        # this is a little bit of a hack for container member devices that
        # need aliases, but even more of a hack for btrfs since you cannot tell
        # from inside the BTRFS class whether you're dealing with a member or a
        # volume/subvolume
        if self.format.type == "btrfs" and not self.type.startswith("btrfs"):
            data.mountpoint = "btrfs."  # continued below, also for lvm, raid

        if data.mountpoint.endswith("."):
            data.mountpoint += str(self.id)

    def is_name_valid(self, name):
        # This device corresponds to a file in /dev, so no /'s or nulls,
        # and the name cannot be . or ..

        # ...except some names *do* contain directory components, for this
        # is an imperfect world of joy and sorrow mingled. For cciss, split
        # the path into its components and do the real check on each piece
        if name.startswith("cciss/"):
            return all(self.is_name_valid(n) for n in name.split('/'))

        badchars = any(c in ('\x00', '/') for c in name)
        return not badchars and name != '.' and name != '..'



# task.py
# Abstract class for tasks.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass


@add_metaclass(abc.ABCMeta)
class Task(object):

    """ An abstract class that represents some task. """

    # Whether or not the functionality is implemented in the task class.
    # It is True by default. Only NotImplementedClass and its descendants
    # should have a False value.
    implemented = True

    description = abc.abstractproperty(doc="Brief description for this task.")

    @property
    def availability_errors(self):
        """ Reasons if this task or the tasks it depends on are unavailable. """
        return self._availability_errors + \
            [e for t in self.depends_on for e in t.availability_errors]

    @property
    def available(self):
        """ True if the task is available, otherwise False.

            :returns: True if the task is available
            :rtype: bool
        """
        return self.availability_errors == []

    _availability_errors = abc.abstractproperty(
        doc="Reasons if the necessary external tools are unavailable.")

    depends_on = abc.abstractproperty(doc="tasks that this task depends on")

    @abc.abstractmethod
    def do_task(self, *args, **kwargs):
        """ Do the task for this class. """
        raise NotImplementedError()


class UnimplementedTask(Task):

    """ A null Task, which returns a negative or empty for all properties."""

    description = "an unimplemented task"
    implemented = False

    @property
    def _availability_errors(self):
        return ["Not implemented task can not succeed."]

    depends_on = []

    def __str__(self):
        return "unimplemented task"

    def do_task(self, *args, **kwargs):
        raise NotImplementedError()


@add_metaclass(abc.ABCMeta)
class BasicApplication(Task):

    """ A task representing an application. """

    ext = abc.abstractproperty(doc="The object representing the external resource.")

    # TASK methods

    def __str__(self):
        return str(self.ext)

    @property
    def _availability_errors(self):
        errors = self.ext.availability_errors
        if errors:
            return ["application %s is not available: %s" % (self.ext, " and ".join(errors))]
        else:
            return []

    @property
    def depends_on(self):
        return []



#!/usr/bin/python3

import setuptools
from setuptools import setup
from setuptools.command.sdist import sdist
import subprocess
import sys
import os

# this is copied straight from distutils.filelist.findall , but with os.stat()
# replaced with os.lstat(), so S_ISLNK() can actually tell us something.


def findall(dirname=os.curdir):
    from stat import ST_MODE, S_ISREG, S_ISDIR, S_ISLNK

    file_list = []
    stack = [dirname]
    pop = stack.pop
    push = stack.append

    while stack:
        dirname = pop()
        names = os.listdir(dirname)

        for name in names:
            if dirname != os.curdir:        # avoid the dreaded "./" syndrome
                fullname = os.path.join(dirname, name)
            else:
                fullname = name

            # Avoid excess stat calls -- just one will do, thank you!
            stat = os.lstat(fullname)
            mode = stat[ST_MODE]
            if S_ISREG(mode):
                file_list.append(fullname)
            elif S_ISDIR(mode) and not S_ISLNK(mode):
                push(fullname)

    return file_list

setuptools.findall = findall

# Extend the sdist command
class blivet_sdist(sdist):
    user_options = sdist.user_options + [('mode=', None, "specify mode for sdist; one of 'release', 'normal'"),]

    def initialize_options(self):
        sdist.initialize_options(self)
        self.mode = None  # pylint: disable=attribute-defined-outside-init

    def finalize_options(self):
        sdist.finalize_options(self)
        if self.mode not in (None, 'release', 'normal'):
            raise AttributeError('Unknown mode %s' % self.mode)

    def run(self):
        # Build the .mo files
        subprocess.check_call(['make', '-C', 'po'])

        # Run the parent command
        sdist.run(self)

    def make_release_tree(self, base_dir, files):
        # Run the parent command first
        sdist.make_release_tree(self, base_dir, files)

        if self.mode == "release":
            # Run translation-canary in release mode to remove any bad translations
            sys.path.append('translation-canary')
            from translation_canary.translated import testSourceTree  # pylint: disable=import-error
            testSourceTree(base_dir, releaseMode=True)


data_files = [
    ('/etc/dbus-1/system.d', ['dbus/blivet.conf']),
    ('/usr/share/dbus-1/system-services', ['dbus/com.redhat.Blivet0.service']),
    ('/usr/libexec', ['dbus/blivetd']),
    ('/usr/lib/systemd/system', ['dbus/blivet.service'])
]


with open("README.md", "r") as f:
    long_description = f.read()


setup(name='blivet',
      version='3.8.2',
      cmdclass={"sdist": blivet_sdist},
      description='Python module for system storage configuration',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='David Lehman', author_email='dlehman@redhat.com',
      url='http://github.com/storaged-project/blivet',
      data_files=data_files,
      packages=['blivet', 'blivet.dbus', 'blivet.devices', 'blivet.devicelibs', 'blivet.events', 'blivet.formats', 'blivet.populator', 'blivet.static_data', 'blivet.tasks', 'blivet.populator.helpers'],
      install_requires=['pyudev', 'six'],
      classifiers=["Development Status :: 5 - Production/Stable",
                   "Intended Audience :: Developers",
                   "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
                   "Programming Language :: Python :: 3",
                   "Operating System :: POSIX :: Linux"]
     )



# volume_info.py
# Backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): Jan Pokorny <japokorn@redhat.com>
#

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")


class LVsInfo(object):
    """ Class to be used as a singleton.
        Maintains the LVs cache.
    """

    def __init__(self):
        self._lvs_cache = None

    @property
    def cache(self):
        if self._lvs_cache is None:
            try:
                lvs = blockdev.lvm.lvs()
            except NotImplementedError:
                log.error("libblockdev lvm plugin is missing")
                self._lvs_cache = dict()
                return self._lvs_cache

            self._lvs_cache = dict(("%s-%s" % (lv.vg_name, lv.lv_name), lv) for lv in lvs)

        return self._lvs_cache

    def drop_cache(self):
        self._lvs_cache = None


lvs_info = LVsInfo()


class PVsInfo(object):
    """ Class to be used as a singleton.
        Maintains the PVs cache.
    """

    def __init__(self):
        self._pvs_cache = None

    @property
    def cache(self):
        if self._pvs_cache is None:
            self._pvs_cache = dict()

            try:
                pvs = blockdev.lvm.pvs()
            except NotImplementedError:
                log.error("libblockdev lvm plugin is missing")
                return self._pvs_cache

            for pv in pvs:
                self._pvs_cache[pv.pv_name] = pv
                # TODO: add get_all_device_symlinks() and resolve_device_symlink() functions to
                #       libblockdev and use them here
                if pv.pv_name.startswith("/dev/md/"):
                    try:
                        md_node = blockdev.md.node_from_name(pv.pv_name[len("/dev/md/"):])
                        self._pvs_cache["/dev/" + md_node] = pv
                    except blockdev.MDRaidError:
                        pass
                elif pv.pv_name.startswith("/dev/md"):
                    try:
                        md_named_dev = blockdev.md.name_from_node(pv.pv_name[len("/dev/"):])
                        self._pvs_cache["/dev/md/" + md_named_dev] = pv
                    except blockdev.MDRaidError:
                        pass

        return self._pvs_cache

    def drop_cache(self):
        self._pvs_cache = None


pvs_info = PVsInfo()


class VGsInfo(object):
    """ Class to be used as a singleton.
        Maintains the VGs cache.
    """

    def __init__(self):
        self._vgs_cache = None

    @property
    def cache(self):
        if self._vgs_cache is None:
            try:
                vgs = blockdev.lvm.vgs()
            except NotImplementedError:
                log.error("libblockdev lvm plugin is missing")
                self._vgs_cache = dict()
                return self._vgs_cache

            self._vgs_cache = dict(("%s" % (vg.uuid), vg) for vg in vgs)

        return self._vgs_cache

    def drop_cache(self):
        self._vgs_cache = None


vgs_info = VGsInfo()



import copy
import unittest

import blivet.formats as formats


class FormatsTestCase(unittest.TestCase):

    def test_formats_methods(self):
        ##
        # get_device_format_class
        ##
        format_pairs = {
            None: formats.DeviceFormat,
            "bogus": None,
            "biosboot": formats.biosboot.BIOSBoot,
            "BIOS Boot": formats.biosboot.BIOSBoot,
            "nodev": formats.fs.NoDevFS
        }
        format_names = format_pairs.keys()
        format_values = [format_pairs[k] for k in format_names]

        self.assertEqual(
            [formats.get_device_format_class(x) for x in format_names],
            format_values)

        # A DeviceFormat object is returned if lookup by name fails
        for name in format_names:
            self.assertIs(formats.get_format(name).__class__,
                          formats.DeviceFormat if format_pairs[name] is None else format_pairs[name])
        # Consecutively constructed DeviceFormat objects have consecutive ids
        names = [key for key in format_pairs.keys() if format_pairs[key] is not None]
        objs = [formats.get_format(name) for name in names]
        ids = [obj.id for obj in objs]
        self.assertEqual(ids, list(range(ids[0], ids[0] + len(ids))))

        # Copy or deepcopy should preserve the id
        self.assertEqual(ids, [copy.copy(obj).id for obj in objs])
        self.assertEqual(ids, [copy.deepcopy(obj).id for obj in objs])



# Entry point for testing translations
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>

import sys, argparse
from . import testSourceTree

ap = argparse.ArgumentParser(description='Validate translated strings')
ap.add_argument('--release', action='store_true', default=False,
        help='Run in release mode')
ap.add_argument('--test', dest='release', action='store_false',
        help='Run in test mode')
ap.add_argument('--no-modify-linguas', dest='modify_linguas', action='store_false', default=True,
        help='In release mode, do not remove failing translation from LINGUAS')
ap.add_argument('source_trees', metavar='SOURCE-TREE', nargs='+',
        help='Source directory to test')

args = ap.parse_args()

status = 0
for srcdir in args.source_trees:
    if not testSourceTree(srcdir, args.release, args.modify_linguas):
        status = 1

sys.exit(status)



import os
import unittest

from ..storagetestcase import StorageTestCase

import blivet

from blivet.devices.stratis import StratisFilesystemDevice


class StratisTestCase(StorageTestCase):

    @classmethod
    def setUpClass(cls):
        unavailable_deps = StratisFilesystemDevice.unavailable_type_dependencies()
        if unavailable_deps:
            dep_str = ", ".join([d.name for d in unavailable_deps])
            raise unittest.SkipTest("some unavailable dependencies required for this test: %s" % dep_str)

    def setUp(self):
        super().setUp()

        disks = [os.path.basename(vdev) for vdev in self.vdevs]
        self.storage = blivet.Blivet()
        self.storage.exclusive_disks = disks
        self.storage.reset()

        # make sure only the targetcli disks are in the devicetree
        for disk in self.storage.disks:
            self.assertTrue(disk.path in self.vdevs)
            self.assertIsNone(disk.format.type)
            self.assertFalse(disk.children)

    def _clean_up(self):
        self.storage.reset()
        for disk in self.storage.disks:
            if disk.path not in self.vdevs:
                raise RuntimeError("Disk %s found in devicetree but not in disks created for tests" % disk.name)
            self.storage.recursive_remove(disk)

        self.storage.do_it()

        return super()._clean_up()

    def test_stratis_basic(self):
        disk = self.storage.devicetree.get_device_by_path(self.vdevs[0])
        self.assertIsNotNone(disk)
        self.storage.initialize_disk(disk)

        bd = self.storage.new_partition(size=blivet.size.Size("1 GiB"), fmt_type="stratis",
                                        parents=[disk])
        self.storage.create_device(bd)

        blivet.partitioning.do_partitioning(self.storage)

        pool = self.storage.new_stratis_pool(name="blivetTestPool", parents=[bd])
        self.storage.create_device(pool)

        fs = self.storage.new_stratis_filesystem(name="blivetTestFS", parents=[pool],
                                                 size=blivet.size.Size("800 MiB"))
        self.storage.create_device(fs)

        self.storage.do_it()
        self.storage.reset()

        disk = self.storage.devicetree.get_device_by_path(self.vdevs[0])
        self.assertIsNotNone(disk)

        bd = self.storage.devicetree.get_device_by_path(self.vdevs[0] + "1")
        self.assertIsNotNone(bd)
        self.assertIsInstance(bd, blivet.devices.PartitionDevice)
        self.assertIsNotNone(bd.format)
        self.assertEqual(bd.format.type, "stratis")

        pool = self.storage.devicetree.get_device_by_name("blivetTestPool")
        self.assertIsNotNone(pool)
        self.assertIsInstance(pool, blivet.devices.StratisPoolDevice)
        self.assertIsNotNone(pool.format)
        self.assertIsNone(pool.format.type)
        self.assertEqual(bd.format.pool_name, pool.name)
        self.assertEqual(len(pool.parents), 1)
        self.assertEqual(pool.parents[0], bd)

        fs = self.storage.devicetree.get_device_by_name("blivetTestPool/blivetTestFS")
        self.assertIsNotNone(fs)
        self.assertIsInstance(fs, blivet.devices.StratisFilesystemDevice)
        self.assertIsNotNone(fs.format)
        self.assertEqual(fs.format.type, "stratis xfs")
        self.assertEqual(fs.pool, pool)
        self.assertEqual(len(fs.parents), 1)
        self.assertEqual(fs.parents[0], pool)

    def test_stratis_encrypted(self):
        disk = self.storage.devicetree.get_device_by_path(self.vdevs[0])
        self.assertIsNotNone(disk)
        self.storage.initialize_disk(disk)

        bd = self.storage.new_partition(size=blivet.size.Size("1 GiB"), fmt_type="stratis",
                                        parents=[disk])
        self.storage.create_device(bd)

        blivet.partitioning.do_partitioning(self.storage)

        pool = self.storage.new_stratis_pool(name="blivetTestPool", parents=[bd],
                                             encrypted=True, passphrase="abcde")
        self.storage.create_device(pool)

        self.storage.do_it()
        self.storage.reset()

        disk = self.storage.devicetree.get_device_by_path(self.vdevs[0])
        self.assertIsNotNone(disk)

        bd = self.storage.devicetree.get_device_by_path(self.vdevs[0] + "1")
        self.assertIsNotNone(bd)
        self.assertIsInstance(bd, blivet.devices.PartitionDevice)
        self.assertIsNotNone(bd.format)
        self.assertEqual(bd.format.type, "stratis")

        pool = self.storage.devicetree.get_device_by_name("blivetTestPool")
        self.assertIsNotNone(pool)
        self.assertIsInstance(pool, blivet.devices.StratisPoolDevice)
        self.assertIsNotNone(pool.format)
        self.assertIsNone(pool.format.type)
        self.assertEqual(bd.format.pool_name, pool.name)
        self.assertEqual(len(pool.parents), 1)
        self.assertEqual(pool.parents[0], bd)
        self.assertTrue(pool.encrypted)

    def test_stratis_overprovision(self):
        disk = self.storage.devicetree.get_device_by_path(self.vdevs[0])
        self.assertIsNotNone(disk)
        self.storage.initialize_disk(disk)

        bd = self.storage.new_partition(size=blivet.size.Size("1 GiB"), fmt_type="stratis",
                                        parents=[disk])
        self.storage.create_device(bd)

        blivet.partitioning.do_partitioning(self.storage)

        pool = self.storage.new_stratis_pool(name="blivetTestPool", parents=[bd])
        self.storage.create_device(pool)

        self.storage.do_it()
        self.storage.reset()

        pool = self.storage.devicetree.get_device_by_name("blivetTestPool")
        self.assertIsNotNone(pool)

        with self.assertRaisesRegex(blivet.errors.StratisError, "not enough free space in the pool"):
            fs = self.storage.new_stratis_filesystem(name="blivetTestFS", parents=[pool],
                                                     size=blivet.size.Size("1 TiB"))

        fs = self.storage.new_stratis_filesystem(name="blivetTestFS", parents=[pool],
                                                 size=blivet.size.Size("2 GiB"))
        self.storage.create_device(fs)

        self.storage.do_it()
        self.storage.reset()

        fs = self.storage.devicetree.get_device_by_name("blivetTestPool/blivetTestFS")
        self.assertIsNotNone(fs)
        self.assertIsInstance(fs, blivet.devices.StratisFilesystemDevice)
        self.assertAlmostEqual(fs.size, blivet.size.Size("2 GiB"), delta=blivet.size.Size("10 MiB"))



from .device_dependencies_test import *
from .device_methods_test import *
from .device_names_test import *
from .device_packages_test import *
from .device_properties_test import *
from .device_size_test import *
from .disk_test import *
from .lvm_test import *
from .md_test import *
from .network_test import *
from .partition_test import *
from .stratis_test import *
